unit gpx;

interface

uses
  SysUtils, Classes, Contnrs;

type

  TFix = (unknow, none, d2, d3, dgps, pps);
  TBound = record
    LatMin, LatMax, LongMin, LongMax : real;
  end;

const

  FixTxt : array[TFix] of string =
           ('', 'none', '2d', '3d', 'dgps', 'pps');

type

  TGpxBasic = class
  private
    FTag: string;
    procedure SetTag(const Value: string);
    procedure GenerateInside(var content : TStringList; indent : string); virtual; abstract;
    function GenerateHeader : string; virtual;
    procedure Generate(var content : TStringList; indent : string); virtual;
    procedure Analyze_XML(var header, content : string; const Strict : boolean;
                                CDataValues : array of string; const CodePage : word;
                                var Error_Msg, Warning_Msg : string;
                                var Succes : boolean); virtual; abstract;
  public
    property Tag : string read FTag write SetTag;
    Constructor Create(const Tag_Value : string); virtual;
  end;

  TGpxLink = class(TGPXBasic) // linkType
  private
    FLinkText: string;
    FURI: string;
    FLinkType: string;
    procedure SetLinkText(const Value: string);
    procedure SetLinkType(const Value: string);
    procedure SetURI(const Value: string);
    procedure GenerateInside(var content : TStringList; indent : string); override;
    function GenerateHeader : string; override;
    procedure Analyze_XML(var header, content : string; const Strict : boolean;
                                CDataValues : array of string; const CodePage : word;
                                var Error_Msg, Warning_Msg : string;
                                var Succes : boolean); override;
  public
    property URI : string read FURI write SetURI;                       // <... href="xsd:anyURI [1] ?">
    property LinkText : string read FLinkText write SetLinkText;        // <text> xsd:string </text> [0..1] ?
    property LinkType : string read FLinkType write SetLinkType;        // <type> xsd:string </type> [0..1] ?
  end;

  TGpxExtension = class(TGpxBasic)
  private
    FExtension_Content: string;
    procedure GenerateInside(var content : TStringList; indent : string); override;
    procedure SetExtension_Content(const Value: string);
    procedure Analyze_XML(var header, content : string; const Strict : boolean;
                                CDataValues : array of string; const CodePage : word;
                                var Error_Msg, Warning_Msg : string;
                                var Succes : boolean); override;
  public
    property Extension_Content : string read FExtension_Content write SetExtension_Content;
  end;

  TGPXBasicExt = class(TGPXBasic) // basic with <extension> field
  private
    FExtension: TGPXExtension;
    procedure SetExtension(const Value: TGPXExtension);
    procedure ReadExtension(NewTag, Newheader, Newcontent : string; const Strict : boolean;
                                CDataValues : array of string; const CodePage : word;
                                var Error_Msg, Warning_Msg : string; var Succes : boolean);
    procedure GenerateInside(var content : TStringList; indent : string); override;
  public
    property Extension : TGPXExtension read FExtension write SetExtension;  // <extensions> extensionsType </extensions> [0..1] ?
    procedure DeleteExtension;
  end;

  TGpxCommon = class(TGpxBasicExt) // common parts to waypoint, route, track and metadata
  private
    FName: string;
    FDescription: string;
    FLink_List : TObjectList;
    procedure SetName(const Value: string);
    procedure SetDescription(const Value: string);
    function GetLinks(Index: Integer): TGpxLink;
    procedure GenerateInside(var content : TStringList; indent : string); override;
    function GetNbLinks: integer;
    procedure ReadName(NewTag, Newheader, Newcontent : string; const Strict : boolean;
                                CDataValues : array of string; const CodePage : word;
                                var Error_Msg, Warning_Msg : string; var Succes : boolean);
    procedure ReadDescription(NewTag, Newheader, Newcontent : string; const Strict : boolean;
                                CDataValues : array of string; const CodePage : word;
                                var Error_Msg, Warning_Msg : string; var Succes : boolean);
    procedure ReadLink(NewTag, Newheader, Newcontent : string; const Strict : boolean;
                                CDataValues : array of string; const CodePage : word;
                                var Error_Msg, Warning_Msg : string; var Succes : boolean);
  public
    // Optional Description Information
    property Name : string read FName write SetName;                            // <name> GPS waypoint name of the waypoint
    property Description : string read FDescription write SetDescription;       // <desc> Descriptive description of the waypoint
    property Links[Index: Integer]: TGpxLink read GetLinks;
    property NbLinks: integer read GetNbLinks;
    procedure AddLink(const Value: TGpxLink);
    procedure DeleteLink(const Index : integer);

    Constructor Create(const Tag_Value : string); override;
    destructor Destroy; override;
  end;

  TGpxCommonData = class(TGpxCommon) // common parts to waypoint, route and track
  private
    FComment: string;
    FSource: string;
    FCategory: string;
    procedure SetComment(const Value: string);
    procedure SetSource(const Value: string);
    procedure SetCategory(const Value: string);
    procedure GenerateInside(var content : TStringList; indent : string); override;
    procedure ReadComment(NewTag, Newheader, Newcontent : string; const Strict : boolean;
                                CDataValues : array of string; const CodePage : word;
                                var Error_Msg, Warning_Msg : string; var Succes : boolean);
    procedure ReadSource(NewTag, Newheader, Newcontent : string; const Strict : boolean;
                                CDataValues : array of string; const CodePage : word;
                                var Error_Msg, Warning_Msg : string; var Succes : boolean);
    procedure ReadCategory(NewTag, Newheader, Newcontent : string; const Strict : boolean;
                                CDataValues : array of string; const CodePage : word;
                                var Error_Msg, Warning_Msg : string; var Succes : boolean);
  public
    // Optional Description Information
    property Comment : string read FComment write SetComment;                   // <cmt> GPS comment of the waypoint
    property Source : string read FSource write SetSource;                      // <src> Source of the waypoint data
    property Category : string read FCategory write SetCategory;                // <type> Type (category) of waypoint
  end;


  TGPXWaypoint = class(TGpxCommonData) // wptType
  private
    FNb_Sat: integer;
    FDGPSid: integer;
    FDGPS_Age: real;
    FLatitude: real;
    FPDOP: real;
    FMagVar: real;
    FHDOP: real;
    FLongitude: real;
    FElevation: real;
    FGeoidHeight: real;
    FVDOP: real;
    FSymbol: string;
    FTime: TDateTime;
    FFix: TFix;
    fDefLatitude: boolean;
    fDefLongitude: boolean;
    fDefElevation: boolean;
    fDefTime: boolean;
    fDefDGPSid: boolean;
    fDefMagVar: boolean;
    fDefHDOP: boolean;
    fDefDGPS_Age: boolean;
    fDefNb_Sat: boolean;
    fDefPDOP: boolean;
    fDefGeoidHeight: boolean;
    fDefVDOP: boolean;
    procedure SetDGPSid(const Value: integer);
    procedure SetDGPS_Age(const Value: real);
    procedure SetElevation(const Value: real);
    procedure SetFix(const Value: TFix);
    procedure SetGeoidHeight(const Value: real);
    procedure SetHDOP(const Value: real);
    procedure SetLatitude(const Value: real);
    procedure SetLongitude(const Value: real);
    procedure SetMagVar(const Value: real);
    procedure SetNb_Sat(const Value: integer);
    procedure SetPDOP(const Value: real);
    procedure SetSymbol(const Value: string);
    procedure SetTime(const Value: TDateTime);
    procedure SetVDOP(const Value: real);
    function GetValid: boolean;
    procedure GenerateInside(var content : TStringList; indent : string); override;
    function GenerateHeader : string; override;
    procedure Analyze_XML(var header, content : string; const Strict : boolean;
                                CDataValues : array of string; const CodePage : word;
                                var Error_Msg, Warning_Msg : string;
                                var Succes : boolean); override;
  public
    // DefXXX indicate if a numeric field is defined
    // Fix as a specific value : unkown
    // string field are just left empty
    // Required Information
    property Latitude : real read FLatitude write SetLatitude;                  // <lat> Latitude of the waypoint.
    property DefLatitude : boolean read fDefLatitude;
    procedure InvalidateLatitude;
    property Longitude : real read FLongitude write SetLongitude;               // <lon> Longitude of the waypoint.
    property DefLongitude : boolean read fDefLongitude;
    procedure InvalidateLongitude;
    // Optional Position Information
    property Elevation : real read FElevation write SetElevation;               // <ele> Elevation of the waypoint.
    property DefElevation : boolean read fDefElevation;
    procedure InvalidateElevation;
    property Time : TDateTime read FTime write SetTime;                         // <time> Creation date/time of the waypoint
    property DefTime : boolean read fDefTime;
    procedure InvalidateTime;
    property MagVar : real read FMagVar write SetMagVar;                        // <magvar> Magnetic variation of the waypoint
    property DefMagVar : boolean read fDefMagVar;
    property GeoidHeight : real read FGeoidHeight write SetGeoidHeight;         // <geoidheight> Geoid height of the waypoint
    property DefGeoidHeight : boolean read fDefGeoidHeight;
    procedure InvalidateGeoidHeight;
    // Optional Description Information
    property Symbol : string read FSymbol write SetSymbol;                      // <sym> Waypoint symbol
    // Optional Accuracy Information
    property Fix : TFix read FFix write SetFix;                                 // <fix> Type of GPS fix
    property Nb_Sat : integer read FNb_Sat write SetNb_Sat;                     // <sat> Number of satellites
    property DefNb_Sat : boolean read fDefNb_Sat;
    procedure InvalidateNb_Sat;
    property HDOP : real read FHDOP write SetHDOP;                              // <hdop> HDOP
    property DefHDOP : boolean read fDefHDOP;
    procedure InvalidateHDOP;
    property VDOP : real read FVDOP write SetVDOP;                              // <vdop> VDOP
    property DefVDOP : boolean read fDefVDOP;
    procedure InvalidateVDOP;
    property PDOP : real read FPDOP write SetPDOP;                              // <pdop> PDOP
    property DefPDOP : boolean read fDefPDOP;
    procedure InvalidatePDOP;
    property DGPS_Age : real read FDGPS_Age write SetDGPS_Age;                  // <ageofdgpsdata> Time since last DGPS fix
    property DefDGPS_Age : boolean read fDefDGPS_Age;
    procedure InvalidateDGPS_Age;
    property DGPSid : integer read FDGPSid write SetDGPSid;                     // <dgpsid> DGPS station ID
    property DefDGPSid : boolean read fDefDGPSid;
    procedure InvalidateDGPSid;

    property Valid : boolean read GetValid;

    Constructor Create(const Tag_Value : string); override;
    destructor Destroy; override;
  end;

  TGPXTrackSegment = class(TGPXBasicExt) // trksegType
  private
    FWayPoints_List : TObjectList;
    function GetWayPoints(Index: Integer): TGPXWaypoint;
    procedure GenerateInside(var content : TStringList; indent : string); override;
    function GetNbWayPoints: integer;
    procedure Analyze_XML(var header, content : string; const Strict : boolean;
                                CDataValues : array of string; const CodePage : word;
                                var Error_Msg, Warning_Msg : string;
                                var Succes : boolean); override;
  public
    property WayPoints[Index: Integer]: TGPXWaypoint read GetWaypoints;         // <trkpt> wptType </trkpt> [0..*] ?
    property NbWaypoints : integer read GetNbWayPoints;
    procedure AddWayPoint(const Value: TGPXWaypoint);
    procedure DeleteWayPoint(const Index : integer);

    Constructor Create(const Tag_Value : string); override;
    destructor Destroy; override;
  end;


  TGPXRoute = class(TGpxCommonData) // rteType
  private
    FWayPoints_List : TObjectList;
    fDefNumber: boolean;
    FNumber: integer;
    function GetWayPoints(Index: Integer): TGPXWaypoint;
    procedure SetNumber(const Value: integer);
    procedure GenerateInside(var content : TStringList; indent : string); override;
    function GetNbWayPoints: integer;
    procedure Analyze_XML(var header, content : string; const Strict : boolean;
                                CDataValues : array of string; const CodePage : word;
                                var Error_Msg, Warning_Msg : string;
                                var Succes : boolean); override;
  public
    property WayPoints[Index: Integer]: TGPXWaypoint read GetWayPoints;         // <rtept> wptType </rtept> [0..*] ?
    property NbWaypoints : integer read GetNbWayPoints;
    procedure AddWayPoint(const Value: TGPXWaypoint);
    procedure DeleteWayPoint(const Index : integer);
    property Number : integer read FNumber write SetNumber;                     // <number> xsd:nonNegativeInteger </number> [0..1] ?
    property DefNumber : boolean read fDefNumber;
    procedure InvalidateNumber;

    Constructor Create(const Tag_Value : string); override;
    destructor Destroy; override;
  end;


  TGPXTrack = class(TGpxCommonData)
  private
    FTrackSegs_List : TObjectList;
    fDefNumber: boolean;
    FNumber: integer;
    function GetTrackSegs(Index: Integer): TGPXTrackSegment;
    procedure GenerateInside(var content : TStringList; indent : string); override;
    function GetNbTrackSegs: integer;
    procedure SetNumber(const Value: integer);
    procedure Analyze_XML(var header, content : string; const Strict : boolean;
                                CDataValues : array of string; const CodePage : word;
                                var Error_Msg, Warning_Msg : string;
                                var Succes : boolean); override;
  public
    property TrackSegs[Index: Integer]: TGPXTrackSegment read GetTrackSegs;
    property NbTrackSegs : integer read GetNbTrackSegs;
    procedure AddTrackSeg(const Value: TGPXTrackSegment);
    procedure DeleteTrackSeg(const Index : integer);
    property Number : integer read FNumber write SetNumber;                     // <number> xsd:nonNegativeInteger </number> [0..1] ?
    property DefNumber : boolean read fDefNumber;
    procedure InvalidateNumber;

    Constructor Create(const Tag_Value : string); override;
    destructor Destroy; override;
  end;

  TGPXPerson =  class(TGPXBasic) // personType
  private
    FEmail_id: string;
    FEmail_domain: string;
    FName: string;
    FLink: TGPXLink;
    procedure SetEmail_domain(const Value: string);
    procedure SetEmail_id(const Value: string);
    procedure SetLink(const Value: TGPXLink);
    procedure SetName(const Value: string);
    procedure GenerateInside(var content : TStringList; indent : string); override;
    procedure Analyze_XML(var header, content : string; const Strict : boolean;
                                CDataValues : array of string; const CodePage : word;
                                var Error_Msg, Warning_Msg : string;
                                var Succes : boolean); override;
  public
    property Name : string read FName write SetName;
    property Email_id : string read FEmail_id write SetEmail_id;
    property Email_domain : string read FEmail_domain write SetEmail_domain;
    property Link : TGPXLink read FLink write SetLink;
    procedure DeleteLink;

    destructor Destroy; override;
  end;

  TGPXCopyright = class(TGPXBasic) // copyrightType
  private
    FDefYear: boolean;
    FYear: integer;
    FAuthor: string;
    FLicense: string;
    procedure SetAuthor(const Value: string);
    procedure SetLicense(const Value: string);
    procedure SetYear(const Value: integer);
    procedure GenerateInside(var content : TStringList; indent : string); override;
    function GenerateHeader : string; override;
    procedure Analyze_XML(var header, content : string; const Strict : boolean;
                                CDataValues : array of string; const CodePage : word;
                                var Error_Msg, Warning_Msg : string;
                                var Succes : boolean); override;
  public
    property Author : string read FAuthor write SetAuthor;      // <...author="xsd:string [1] ?">
    property Year : integer read FYear write SetYear;           // <year> xsd:gYear </year> [0..1] ?
    property DefYear : boolean read FDefYear;
    procedure InvalidateYear;
    property License : string read FLicense write SetLicense;   // <license> xsd:anyURI </license> [0..1] ?

    Constructor Create(const Tag_Value : string); override;
  end;



  TGPXMetaData = class(TGPXCommon) //  metadataType
  private
    FDefBounds: boolean;
    FDefTime: boolean;
    FKeywords: string;
    FBounds: TBound;
    FTime: TDateTime;
    FCopyright: TGPXCopyright;
    FAuthor: TGPXPerson;
    procedure SetAuthor(const Value: TGPXPerson);
    procedure SetBounds(const Value: TBound);
    procedure SetCopyright(const Value: TGPXCopyright);
    procedure SetKeywords(const Value: string);
    procedure SetTime(const Value: TDateTime);
    procedure GenerateInside(var content : TStringList; indent : string); override;
    procedure Analyze_XML(var header, content : string; const Strict : boolean;
                                CDataValues : array of string; const CodePage : word;
                                var Error_Msg, Warning_Msg : string;
                                var Succes : boolean); override;
  public
    property Author : TGPXPerson read FAuthor write SetAuthor;                  // <author> personType </author> [0..1] ?
    procedure DeleteAuthor;
    property Copyright : TGPXCopyright read FCopyright write SetCopyright;      // <copyright> copyrightType </copyright> [0..1] ?
    procedure DeleteCopyright;
    property Time : TDateTime read FTime write SetTime;                         // <time> xsd:dateTime </time> [0..1] ?
    property DefTime : boolean read FDefTime;
    procedure InvalidateTime;
    property Keywords : string read FKeywords write SetKeywords;                // <keywords> xsd:string </keywords> [0..1] ?
    property Bounds : TBound read FBounds write SetBounds;                      // <bounds> boundsType </bounds> [0..1] ?
    property DefBounds : boolean read FDefBounds;
    procedure InvalidateBounds;

    Constructor Create(const Tag_Value : string); override;
    destructor Destroy; override;
  end;

  TGPXData = class(TGPXBasicExt) // gpxType, GPX File Organization
  private
    FWayPoints_List : TObjectList;
    FRoute_List : TObjectList;
    FTrack_List : TObjectList;
    FCreator: string;
    FMetadata: TGPXMetaData;
    function GetWayPoints(Index: Integer): TGPXWaypoint;
    procedure SetCreator(const Value: string);
    procedure SetMetadata(const Value: TGPXMetaData);
    function GetRoutes(Index: Integer): TGPXRoute;
    function GetTracks(Index: Integer): TGPXTrack;
    function GetNbRoutes: integer;
    function GetNbTracks: integer;
    function GetNbWaypoints: integer;
    procedure GenerateInside(var content : TStringList; indent : string); override;
    function GenerateHeader : string; override;
    procedure Analyze_XML(var header, content : string; const Strict : boolean;
                                CDataValues : array of string; const CodePage : word;
                                var Error_Msg, Warning_Msg : string;
                                var Succes : boolean); override;
  public
    property Creator : string read FCreator write SetCreator;           //  creator="xsd:string [1] ?">
    property Metadata : TGPXMetaData read FMetadata write SetMetadata;  // <metadata> metadataType </metadata> [0..1] ?
    procedure DeleteMetadata;
    property WayPoints[Index: Integer]: TGPXWaypoint read GetWayPoints; // <wpt> wptType </wpt> [0..*] ?
    property NbWayPoints : integer read GetNbWaypoints;
    procedure AddWayPoint(const Value: TGPXWaypoint);
    procedure DeleteWayPoint (const Index : integer);
    property Routes[Index: Integer]: TGPXRoute read GetRoutes;          // <rte> rteType </rte> [0..*] ?
    property NbRoutes : integer read GetNbRoutes;
    procedure AddRoute(const Value: TGPXRoute);
    procedure DeleteRoute(const Index : integer);
    property Tracks[Index: Integer]: TGPXTrack read GetTracks;          // <trk> trkType </trk> [0..*] ?
    property NbTracks : integer read GetNbTracks;
    procedure AddTrack(const Value: TGPXTrack);
    procedure DeleteTrack(const Index : integer);

    Constructor Create(const Creator_Value : string); override;
    destructor Destroy; override;
    procedure Clear;
    procedure Generate(var content : TStringList); reintroduce;
    procedure SaveToFile(const FileName: string);
    procedure LoadFromFile(const FileName: string; const Strict : boolean;
                                var Error_Msg, Warning_Msg : string;
                                var Succes : boolean);
  end;

implementation

uses
  DateUtils, Math,
  ALUnicode,
  xml_utils, gpx_translation;


{ TGPXWaypoint }

constructor TGPXWaypoint.Create;
begin
  inherited;
  fDefLatitude:= false;
  fDefLongitude:= false;
  fDefElevation:= false;
  fDefTime:= false;
  fDefDGPSid:= false;
  fDefMagVar:= false;
  fDefHDOP:= false;
  fDefDGPS_Age:= false;
  fDefNb_Sat:= false;
  fDefPDOP:= false;
  fDefGeoidHeight:= false;
  fDefVDOP:= false;
  fFix:=unknow;
end;

destructor TGPXWaypoint.Destroy;
begin

  inherited;
end;

function TGPXWaypoint.GetValid: boolean;
begin
  result:=fDefLatitude and fDefLongitude;
end;

procedure TGPXWaypoint.SetDGPSid(const Value: integer);
begin
  FDGPSid := Value;
  fDEFDGPSid := true;
end;

procedure TGPXWaypoint.SetDGPS_Age(const Value: real);
begin
  FDGPS_Age := Value;
  FDefDGPS_Age := true;
end;

procedure TGPXWaypoint.SetElevation(const Value: real);
begin
  FElevation := Value;
  FDefElevation := true;
end;

procedure TGPXWaypoint.SetFix(const Value: TFix);
begin
  FFix := Value;
end;

procedure TGPXWaypoint.SetGeoidHeight(const Value: real);
begin
  FGeoidHeight := Value;
  FDefGeoidHeight := true;
end;

procedure TGPXWaypoint.SetHDOP(const Value: real);
begin
  FHDOP := Value;
  FDefHDOP := true;
end;

procedure TGPXWaypoint.SetLatitude(const Value: real);
begin
  FLatitude := Value;
  FDefLatitude := true;
end;

procedure TGPXWaypoint.SetLongitude(const Value: real);
begin
  FLongitude := Value;
  FDefLongitude := true;
end;

procedure TGPXWaypoint.SetMagVar(const Value: real);
begin
  FMagVar := Value;
  FDefMagVar := true;
end;

procedure TGPXWaypoint.SetNb_Sat(const Value: integer);
begin
  FNb_Sat := Value;
  FDefNb_Sat := true;
end;

procedure TGPXWaypoint.SetPDOP(const Value: real);
begin
  FPDOP := Value;
  FDefPDOP := true;
end;

procedure TGPXWaypoint.SetSymbol(const Value: string);
begin
  FSymbol := Value;
end;

procedure TGPXWaypoint.SetTime(const Value: TDateTime);
begin
  FTime := Value;
  FDefTime := true;
end;

procedure TGPXWaypoint.SetVDOP(const Value: real);
begin
  FVDOP := Value;
  FDefVDOP := true;
end;

procedure TGPXWaypoint.GenerateInside(var content: TStringList;
  indent: string);
begin
  inherited;
  if DefElevation then
    content.Add(StringReplace(Format('%s<ele>%.3f</ele>',[indent,Elevation]),',','.',[rfReplaceAll]));
  if DefTime then
    content.Add(Format('%s<time>%.4d-%.2d-%.2dT%.2d:%.2d:%.2dZ</time>',
                       [indent, YearOf(Time), MonthOf(Time), DayOf(Time),
                       HourOf(Time), MinuteOf(Time), SecondOf(Time) ]));
  if DefMagVar then
    content.Add(StringReplace(Format('%s<magvar>%.2f</magvar>',[indent,MagVar]),',','.',[rfReplaceAll]));
  if DefGeoidHeight then
    content.Add(StringReplace(Format('%s<geoidheight>%.3f</geoidheight>',[indent,GeoidHeight]),',','.',[rfReplaceAll]));
  if Symbol<>'' then
    content.Add(Format('%s<sym>%s</sym>',[indent,ProtectXML(Symbol)]));
  if Fix<>unknow then
    content.Add(Format('%s<fix>%s</fix>',[indent,FixTxt[Fix]]));
  if DefNb_Sat then
    content.Add(Format('%s<sat>%d</sat>',[indent,Nb_Sat]));
  if DefHDOP then
    content.Add(StringReplace(Format('%s<hdop>%.3f</hdop>',[indent,HDOP]),',','.',[rfReplaceAll]));
  if DefVDOP then
    content.Add(StringReplace(Format('%s<vdop>%.3f</vdop>',[indent,VDOP]),',','.',[rfReplaceAll]));
  if DefPDOP then
    content.Add(StringReplace(Format('%s<pdop>%.3f</pdop>',[indent,PDOP]),',','.',[rfReplaceAll]));
  if DefDGPS_Age then
    content.Add(StringReplace(Format('%s<ageofdgpsdata>%.f</ageofdgpsdata>',[indent,DGPS_Age]),',','.',[rfReplaceAll]));
  if DefDGPSid then
    content.Add(Format('%s<dgpsid>%.d</dgpsid>',[indent,DGPSid]));
end;

procedure TGPXWaypoint.InvalidateDGPS_Age;
begin
  FDefDGPS_Age:=false;
end;

procedure TGPXWaypoint.InvalidateDGPSid;
begin
  FDefDGPSid:=false;
end;

procedure TGPXWaypoint.InvalidateElevation;
begin
  FDefElevation:=false;
end;

procedure TGPXWaypoint.InvalidateGeoidHeight;
begin
  FDefGeoidHeight:=false;
end;

procedure TGPXWaypoint.InvalidateHDOP;
begin
  FDefHDOP:=false;
end;

procedure TGPXWaypoint.InvalidateLatitude;
begin
  FDefLatitude:=false;
end;

procedure TGPXWaypoint.InvalidateLongitude;
begin
  FDefLongitude:=false;
end;

procedure TGPXWaypoint.InvalidateNb_Sat;
begin
  FDefNb_Sat:=false;
end;

procedure TGPXWaypoint.InvalidatePDOP;
begin
  FDefPDOP:=false;
end;

procedure TGPXWaypoint.InvalidateTime;
begin
  FDefTime:=false;
end;

procedure TGPXWaypoint.InvalidateVDOP;
begin
  FDefVDOP:=false;
end;

function TGPXWaypoint.GenerateHeader: string;
begin
  result:= StringReplace(Format(' lat="%.8f" lon="%.8f"',[Latitude,Longitude]),',','.',[rfReplaceAll]);
end;

procedure TGPXWaypoint.Analyze_XML(var header, content: string;
  const Strict: boolean; CDataValues : array of string; const CodePage : word;
  var Error_Msg, Warning_Msg: string;
  var Succes: boolean);
var
  NewTag, NewHeader, NewContent, HName, Value : string;
  Empty : boolean;
  Lat_found, Long_found : boolean;
  Angle : real;
  Erreur : integer;
  TheTime : TDateTime;
begin
  Lat_found :=false; Long_found:=false;
  // header analyze
  repeat
    GetFirstValue(header, HName, Value, Empty, Succes);
    ReplaceCData(Value, CDataValues);
    if not Succes then
      Error_Msg:=Format(MsgInvalidElementHeader,[Tag])
    else
    begin
      if not Empty then
      begin
        Case StringIndex(HName, ['lat', 'lon']) of
          0 : begin // lat
            if Lat_Found then // duplicate
              Error_handling(Strict, Error_Msg, Warning_Msg, Succes,
                               MsgDuplicateField,[HName,Tag]);
            if Succes then
            begin
              val(Value,Angle,Erreur);
              if Erreur<>0 then
              begin
                Succes:=false;
                Error_Msg:=Format(MsgInvalidAngle,[Value,Tag]);
              end
              else
              begin
                Lat_found:=true;
                Latitude:=Angle;
              end;
            end;
          end;
          1 : begin // lon
            if long_Found then // duplicate
              Error_handling(Strict, Error_Msg, Warning_Msg, Succes,
                               MsgDuplicateField,[HName,Tag]);
            if Succes then
            begin
              val(Value,Angle,Erreur);
              if Erreur<>0 then
              begin
                Succes:=false;
                Error_Msg:=Format(MsgInvalidAngle,[Value,Tag]);
              end
              else
              begin
                long_found:=true;
                longitude:=Angle;
              end;
            end;
          end;
          else
            Error_handling(Strict, Error_Msg, Warning_Msg, Succes,
                               MsgUnknownField,[HName, Tag]);
        end;
      end;
    end;
  until (not Succes) or empty;

  if Succes and not Lat_Found then
    Error_handling(Strict, Error_Msg, Warning_Msg, Succes,
                               MsgMissingField,['lat',Tag]);

  if Succes and not Long_Found then
    Error_handling(Strict, Error_Msg, Warning_Msg, Succes,
                               MsgMissingField,['lon',Tag]);
  // content analyze
  if Succes then
  begin
    repeat
      GetFirstElement(content, NewTag, NewHeader, NewContent, Empty, Succes);
      if not Succes then
        Error_Msg := Format(MsgInvalidContent,[Tag])
      else
      begin
        if not Empty then
        begin
          Case StringIndex(NewTag, ['name', 'desc', 'cmt', 'src', 'link', 'type',
              'sym', 'ele', 'time', 'magvar', 'geoidheight', 'fix', 'sat', 'hdop',
              'vdop', 'pdop', 'ageofdgpsdata', 'dgpsid', 'extensions']) of
            0 : ReadName(NewTag, NewHeader, NewContent, Strict, CDataValues,
                                CodePage, Error_Msg, Warning_Msg, Succes); // name
            1 : ReadDescription(NewTag, NewHeader, NewContent, Strict, CDataValues,
                                CodePage, Error_Msg, Warning_Msg, Succes); // desc
            2 : ReadComment(NewTag, NewHeader, NewContent, Strict, CDataValues,
                                CodePage, Error_Msg, Warning_Msg, Succes); // cmt
            3 : ReadSource(NewTag, NewHeader, NewContent, Strict, CDataValues,
                                CodePage, Error_Msg, Warning_Msg, Succes); // src
            4 : ReadLink(NewTag, NewHeader, NewContent, Strict, CDataValues, CodePage, Error_Msg, Warning_Msg, Succes); // link
            5 : ReadCategory(NewTag, NewHeader, NewContent, Strict, CDataValues,
                                CodePage, Error_Msg, Warning_Msg, Succes); // type
            6 : begin // sym
              ReplaceCData(NewContent, CDataValues);
              if Symbol<>'' then // duplicate
                Error_handling(Strict, Error_Msg, Warning_Msg, Succes,
                                 MsgDuplicateElement,[NewTag,Tag]);
              if Succes then
                Symbol:=ALUTF8Encode(NewContent, CodePage);
            end;
            7 : begin // ele
              ReplaceCData(NewContent, CDataValues);
              if DefElevation then // duplicate
                Error_handling(Strict, Error_Msg, Warning_Msg, Succes,
                                 MsgDuplicateElement,[NewTag,Tag]);
              if Succes then
              begin
                val(NewContent, Angle, Erreur);
                if Erreur<>0 then
                begin
                  Succes:=false;
                  Error_Msg:=Format(MsgInvalidAngle,[NewContent,NewTag]);
                end
                else
                  Elevation:=Angle;
              end;
            end;
            8 : begin // time
              ReplaceCData(NewContent, CDataValues);
              GetTimeISO8601(NewContent, TheTime, Succes);
              if Succes then
                Time:=TheTime
              else
                Error_Msg:=Format(MsgInvalidTime,[NewContent,Tag]);
            end;
            9 : begin // magvar
              ReplaceCData(NewContent, CDataValues);
              if DefMagVar then // duplicate
                Error_handling(Strict, Error_Msg, Warning_Msg, Succes,
                                 MsgDuplicateElement,[NewTag,Tag]);
              if Succes then
              begin
                val(NewContent, Angle, Erreur);
                if Erreur<>0 then
                begin
                  Succes:=false;
                  Error_Msg:=Format(MsgInvalidAngle,[NewContent,NewTag]);
                end
                else
                  MagVar:=Angle;
              end;
            end;
            10: begin // geoidheight
              ReplaceCData(NewContent, CDataValues);
              if DefGeoidheight then // duplicate
                Error_handling(Strict, Error_Msg, Warning_Msg, Succes,
                                 MsgDuplicateElement,[NewTag,Tag]);
              if Succes then
              begin
                val(NewContent, Angle, Erreur);
                if Erreur<>0 then
                begin
                  Succes:=false;
                  Error_Msg:=Format(MsgInvalidAngle,[NewContent,NewTag]);
                end
                else
                  Geoidheight:=Angle;
              end;
            end;
            11: begin // fix
              ReplaceCData(NewContent, CDataValues);
              if Fix<>unknow then // duplicate
                Error_handling(Strict, Error_Msg, Warning_Msg, Succes,
                                 MsgDuplicateElement,[NewTag,Tag]);
              if Succes then
              begin
                case StringIndex(NewContent,FixTxt) of
                  0 : begin
                    Succes:=false;
                    Error_Msg:=Format(MsgMissingContent,[NewTag]);
                  end;
                  1 : Fix:=none;
                  2 : Fix:=d2;
                  3 : Fix:=d3;
                  4 : Fix:=dgps;
                  5 : Fix:=pps;
                  else
                  begin
                    Succes:=false;
                    Error_Msg:=Format(MsgInvalidValue, [NewTag,NewContent]);
                  end;
                end;
              end;
            end;
            12: begin // sat
              ReplaceCData(NewContent, CDataValues);
              if DefNb_Sat then // duplicate
                Error_handling(Strict, Error_Msg, Warning_Msg, Succes,
                                  MsgDuplicateElement,[NewTag, Tag]);
              if Succes then
              begin
                try
                  Nb_Sat:=StrToInt(NewContent);
                except
                  Succes:=false;
                  Error_Msg:=Format(MsgInvalidValue,[NewTag, NewContent]);
                end;
              end;
            end;
            13: begin // hdop
              ReplaceCData(NewContent, CDataValues);
              if DefHDOP then // duplicate
                Error_handling(Strict, Error_Msg, Warning_Msg, Succes,
                                 MsgDuplicateElement,[NewTag,Tag]);
              if Succes then
              begin
                val(NewContent, Angle, Erreur);
                if Erreur<>0 then
                begin
                  Succes:=false;
                  Error_Msg:=Format(MsgInvalidAngle,[NewContent,NewTag]);
                end
                else
                  HDOP:=Angle;
              end;
            end;
            14: begin // vdop
              ReplaceCData(NewContent, CDataValues);
              if DefVDOP then // duplicate
                Error_handling(Strict, Error_Msg, Warning_Msg, Succes,
                                 MsgDuplicateElement,[NewTag,Tag]);
              if Succes then
              begin
                val(NewContent, Angle, Erreur);
                if Erreur<>0 then
                begin
                  Succes:=false;
                  Error_Msg:=Format(MsgInvalidAngle,[NewContent,NewTag]);
                end
                else
                  VDOP:=Angle;
              end;
            end;
            15: begin // pdop
              ReplaceCData(NewContent, CDataValues);
              if DefPDOP then // duplicate
                Error_handling(Strict, Error_Msg, Warning_Msg, Succes,
                                 MsgDuplicateElement,[NewTag,Tag]);
              if Succes then
              begin
                val(NewContent, Angle, Erreur);
                if Erreur<>0 then
                begin
                  Succes:=false;
                  Error_Msg:=Format(MsgInvalidAngle,[NewContent,NewTag]);
                end
                else
                  PDOP:=Angle;
              end;
            end;
            16: begin // ageofdgpsdata
              ReplaceCData(NewContent, CDataValues);
              if DefDGPS_Age then // duplicate
                Error_handling(Strict, Error_Msg, Warning_Msg, Succes,
                                 MsgDuplicateElement,[NewTag,Tag]);
              if Succes then
              begin
                val(NewContent, Angle, Erreur);
                if Erreur<>0 then
                begin
                  Succes:=false;
                  Error_Msg:=Format(MsgInvalidAngle,[NewContent,NewTag]);
                end
                else
                  DGPS_Age:=Angle;
              end;
            end;
            17: begin // dgpsid
              ReplaceCData(NewContent, CDataValues);
              if DefDGPSid then // duplicate
                Error_handling(Strict, Error_Msg, Warning_Msg, Succes,
                                  MsgDuplicateElement,[NewTag, Tag]);
              if Succes then
              begin
                try
                  DGPSid:=StrToInt(NewContent);
                except
                  Succes:=false;
                  Error_Msg:=Format(MsgInvalidValue,[NewTag, NewContent]);
                end;
              end;
            end;
            18: ReadExtension(NewTag, NewHeader, NewContent, Strict, CDataValues, CodePage,
                                           Error_Msg, Warning_Msg, Succes); // extensions
            else
              Error_handling(Strict, Error_Msg, Warning_Msg, Succes,
                                  MsgUnkownElement,[NewTag, Tag]);
          end;
        end;
      end;
    until (not succes) or empty;
  end;
end;

{ TGpxLink }

procedure TGpxLink.Analyze_XML(var header, content: string;
  const Strict: boolean; CDataValues : array of string; const CodePage : word;
  var Error_Msg, Warning_Msg: string;
  var Succes: boolean);
var
  Empty, HRef_found : boolean;
  Name, Value : string;
  NewTag, NewHeader, NewContent : string;
begin
  Succes:=true;
  HRef_Found:=false;
  // header analyze
  repeat
    GetFirstValue(header, Name, Value, Empty, Succes);
    ReplaceCData(Value, CDataValues);
    if not Succes then
      Error_Msg:=Format(MsgInvalidElementHeader,[Tag])
    else
    begin
      if not Empty then
      begin
        if Name='href' then
        begin
          if HRef_Found then // duplicate
            Error_handling(Strict, Error_Msg, Warning_Msg, Succes,
                               MsgDuplicateField,[Name,Tag]);
          if Succes then
          begin
            URI:=ALUTF8Encode(Value, CodePage);
            HRef_Found:=true;
          end;
        end
        else
          Error_handling(Strict, Error_Msg, Warning_Msg, Succes,
                               MsgUnknownField,[Name,Tag]);
      end;
    end;
  until (not Succes) or empty;

  if Succes and not HRef_Found then
  begin
    Succes:=false;
    Error_Msg:=Format(MsgMissingField,['href',Tag]);
  end;

  // content analyze
  if Succes then
  begin
    repeat
      GetFirstElement(content, NewTag, NewHeader, NewContent, Empty, Succes);
      if not Succes then
        Error_Msg := Format(MsgInvalidContent,[Tag])
      else
      begin
        if not Empty then
        begin
          Case StringIndex(NewTag, ['text', 'type']) of
            0 : begin // text
              ReplaceCData(NewContent, CDataValues);
              if LinkText<>'' then
                Error_handling(Strict, Error_Msg, Warning_Msg, Succes,
                               MsgDuplicateElement,[NewTag,Tag]);
              if Succes then
                LinkText:=ALUTF8Encode(NewContent, CodePage);
            end;
            1 : begin // type
              ReplaceCData(NewContent, CDataValues);
              if LinkType<>'' then
                Error_handling(Strict, Error_Msg, Warning_Msg, Succes,
                               MsgDuplicateElement,[NewTag,Tag]);
              if Succes then
                LinkType:=ALUTF8Encode(NewContent, CodePage);
            end;
            else
              Error_handling(Strict, Error_Msg, Warning_Msg, Succes,
                               MsgUnkownElement,[NewTag, Tag]);
          end;
        end;
      end;
    until (not succes) or empty;
  end;
end;

function TGpxLink.GenerateHeader: string;
begin
  result:=format(' href=%s',[QuoteXMLString(URI)]);
end;

procedure TGpxLink.GenerateInside(var content: TStringList;
  indent: string);
begin
  inherited;
  if LinkText<>'' then
    content.Add(Format('%s<text>%s</text>',[indent,ProtectXML(LinkText)]));
  if LinkType<>'' then
    content.Add(Format('%s<type>%s</type>',[indent,ProtectXML(LinkType)]));
end;

procedure TGpxLink.SetLinkText(const Value: string);
begin
  FLinkText := Value;
end;

procedure TGpxLink.SetLinkType(const Value: string);
begin
  FLinkType := Value;
end;

procedure TGpxLink.SetURI(const Value: string);
begin
  FURI := Value;
end;

{ TGpxCommon }

procedure TGpxCommon.AddLink(const Value: TGpxLink);
begin
  FLink_List.Add(Value);
end;

constructor TGpxCommon.Create;
begin
  inherited;
  FLink_List:=TObjectList.Create;
end;

destructor TGpxCommon.Destroy;
begin
  FLink_List.Free;
  inherited;
end;

procedure TGpxCommon.GenerateInside(var content: TStringList; indent: string);
var
  I : integer;
begin
  inherited;
  if Name<>'' then
    content.Add(Format('%s<name>%s</name>',[indent,ProtectXML(Name)]));
  if Description<>'' then
    content.Add(Format('%s<desc>%s</desc>',[indent,ProtectXML(Description)]));
  for I:=0 to NbLinks-1 do
    TGPXLink(FLink_List[I]).Generate(content,indent);
end;

function TGpxCommon.GetLinks(Index: Integer): TGpxLink;
begin
  assert((Index>=0) and (Index<FLink_List.Count),'Index out of limits in TGpxCommon.GetLinks');
//  assert(Index in [0..FLink_List.Count-1],'Index out of limits in TGpxCommon.GetLinks');
  result:=FLink_List[Index] as TGpxLink;
end;

function TGpxCommon.GetNbLinks: integer;
begin
  result:=FLink_List.Count;
end;

procedure TGpxCommon.DeleteLink(const Index: integer);
begin
  assert((Index>=0) and (Index<FLink_List.Count),'Index out of limits in TGpxCommon.DeleteLink');
//  assert(Index in [0..FLink_List.Count-1],'Index out of limits in TGpxCommon.DeleteLink');
  FLink_List.Delete(Index);
end;

procedure TGpxCommon.SetDescription(const Value: string);
begin
  FDescription := Value;
end;

procedure TGpxCommon.SetName(const Value: string);
begin
  FName := Value;
end;


procedure TGpxCommon.ReadDescription(NewTag, Newheader, Newcontent: string;
  const Strict: boolean; CDataValues : array of string; const CodePage : word;
  var Error_Msg, Warning_Msg: string;
  var Succes: boolean);
begin
  ReplaceCData(NewContent, CDataValues);
  if NewHeader<>'' then // header not empty
    Error_handling(Strict, Error_Msg, Warning_Msg, Succes,
                               MsgUnwantedElementHeader,[NewTag,Tag]);
  if Succes and (Description<>'') then // duplicate
    Error_handling(Strict, Error_Msg, Warning_Msg, Succes,
                               MsgDuplicateElement,[NewTag,Tag]);
  if Succes then
    Description:=ALUTF8Encode(NewContent,CodePage);
end;

procedure TGpxCommon.ReadLink(NewTag, Newheader, Newcontent: string;
  const Strict: boolean; CDataValues : array of string; const CodePage : word;
  var Error_Msg, Warning_Msg: string;
  var Succes: boolean);
var
  Link : TGPXLink;
begin
  Link := TGPXLink.Create(NewTag);
  Link.Analyze_XML(Newheader, Newcontent, Strict, CDataValues, CodePage,
                            Error_Msg, Warning_Msg, Succes);
  if Succes then
    AddLink(Link)
  else
    Link.Free;
end;

procedure TGpxCommon.ReadName(NewTag, Newheader, Newcontent: string;
  const Strict: boolean; CDataValues : array of string; const CodePage : word;
  var Error_Msg, Warning_Msg: string;
  var Succes: boolean);
begin
  ReplaceCData(NewContent, CDataValues);
  if NewHeader<>'' then // header not empty
    Error_handling(Strict, Error_Msg, Warning_Msg, Succes,
                               MsgUnwantedElementHeader,[NewTag,Tag]);
  if Succes and (Name<>'') then // duplicate
    Error_handling(Strict, Error_Msg, Warning_Msg, Succes,
                               MsgDuplicateElement,[NewTag,Tag]);
  if Succes then
    Name:=ALUTF8Encode(NewContent, CodePage);
end;

{ TGpxCommonData }

procedure TGpxCommonData.GenerateInside(var content: TStringList; indent: string);
begin
  inherited;
  if Comment<>'' then
    content.Add(Format('%s<cmt>%s</cmt>',[indent,ProtectXML(Comment)]));
  if Source<>'' then
    content.Add(Format('%s<src>%s</src>',[indent,ProtectXML(Source)]));
  if Category<>'' then
    content.Add(Format('%s<type>%s</type>',[indent,ProtectXML(Category)]));
end;

procedure TGpxCommonData.ReadCategory(NewTag, Newheader,
  Newcontent: string; const Strict: boolean; CDataValues : array of string;
  const CodePage : word;
  var Error_Msg, Warning_Msg: string; var Succes: boolean);
begin
  ReplaceCData(NewContent, CDataValues);
  if NewHeader<>'' then // header not empty
    Error_handling(Strict, Error_Msg, Warning_Msg, Succes,
                               MsgUnwantedElementHeader,[NewTag,Tag]);
  if Succes and (Category<>'') then // duplicate
    Error_handling(Strict, Error_Msg, Warning_Msg, Succes,
                               MsgDuplicateElement,[NewTag,Tag]);
  if Succes then
    Category:=ALUTF8Encode(NewContent, CodePage);
end;

procedure TGpxCommonData.ReadComment(NewTag, Newheader, Newcontent: string;
  const Strict: boolean; CDataValues : array of string; const CodePage : word;
  var Error_Msg, Warning_Msg: string;
  var Succes: boolean);
begin
  ReplaceCData(NewContent, CDataValues);
  if NewHeader<>'' then // header not empty
    Error_handling(Strict, Error_Msg, Warning_Msg, Succes,
                               MsgUnwantedElementHeader,[NewTag,Tag]);
  if Succes and (Comment<>'') then // duplicate
    Error_handling(Strict, Error_Msg, Warning_Msg, Succes,
                               MsgDuplicateElement,[NewTag,Tag]);
  if Succes then
    Comment:=ALUTF8Encode(NewContent, CodePage);
end;

procedure TGpxCommonData.ReadSource(NewTag, Newheader, Newcontent: string;
  const Strict: boolean; CDataValues : array of string; const CodePage : word;
  var Error_Msg, Warning_Msg: string;
  var Succes: boolean);
begin
  ReplaceCData(NewContent, CDataValues);
  if NewHeader<>'' then // header not empty
    Error_handling(Strict, Error_Msg, Warning_Msg, Succes,
                               MsgUnwantedElementHeader,[NewTag,Tag]);
  if Succes and (Source<>'') then // duplicate
    Error_handling(Strict, Error_Msg, Warning_Msg, Succes,
                               MsgDuplicateElement,[NewTag,Tag]);
  if Succes then
    Source:=ALUTF8Encode(NewContent, CodePage);
end;

procedure TGpxCommonData.SetCategory(const Value: string);
begin
  FCategory := Value;
end;

procedure TGpxCommonData.SetComment(const Value: string);
begin
  FComment := Value;
end;

procedure TGpxCommonData.SetSource(const Value: string);
begin
  FSource := Value;
end;

{ TGPXRoute }

procedure TGPXRoute.AddWayPoint(const Value: TGPXWayPoint);
begin
  FWayPoints_List.Add(Value);
end;

procedure TGPXRoute.Analyze_XML(var header, content: string;
  const Strict: boolean; CDataValues : array of string; const CodePage : word;
  var Error_Msg, Warning_Msg: string;
  var Succes: boolean);
var
  NewTag, NewHeader, NewContent : string;
  Empty : boolean;
  WayPoint : TGPXWayPoint;
begin
  if Header<>'' then
    Error_handling(Strict, Error_Msg, Warning_Msg, Succes,
                        MsgUnwantedElementHeader, [Tag,Header]);
  // content analyze
  if Succes then
  begin
    repeat
      GetFirstElement(content, NewTag, NewHeader, NewContent, Empty, Succes);
      if not Succes then
        Error_Msg := Format(MsgInvalidContent,[Tag])
      else
      begin
        if not Empty then
        begin
          Case StringIndex(NewTag, ['name', 'desc', 'cmt', 'src', 'link',
                                  'type', 'number', 'rtept', 'extensions']) of
            0 : ReadName(NewTag, NewHeader, NewContent, Strict, CDataValues,
                                CodePage, Error_Msg, Warning_Msg, Succes); // name
            1 : ReadDescription(NewTag, NewHeader, NewContent, Strict, CDataValues,
                                CodePage, Error_Msg, Warning_Msg, Succes); // desc
            2 : ReadComment(NewTag, NewHeader, NewContent, Strict, CDataValues,
                                CodePage, Error_Msg, Warning_Msg, Succes); // cmt
            3 : ReadSource(NewTag, NewHeader, NewContent, Strict, CDataValues,
                                CodePage, Error_Msg, Warning_Msg, Succes); // src
            4 : ReadLink(NewTag, NewHeader, NewContent, Strict, CDataValues, CodePage, Error_Msg, Warning_Msg, Succes); // link
            5 : ReadCategory(NewTag, NewHeader, NewContent, Strict, CDataValues,
                                CodePage, Error_Msg, Warning_Msg, Succes); // type
            6 : begin // number
              ReplaceCData(NewContent, CDataValues);
              if DefNumber then // duplicate
                Error_handling(Strict, Error_Msg, Warning_Msg, Succes,
                                  MsgDuplicateElement,[NewTag, Tag]);
              if Succes then
              begin
                try
                  Number:=StrToInt(NewContent);
                except
                  Succes:=false;
                  Error_Msg:=Format(MsgInvalidValue,[NewTag, NewContent]);
                end;
              end;
            end;
            7 : begin // rtept
              WayPoint := TGPXWayPoint.Create(NewTag);
              WayPoint.Analyze_XML(Newheader, Newcontent, Strict, CDataValues,
                           CodePage, Error_Msg, Warning_Msg, Succes);
              if Succes then
                AddWayPoint(WayPoint)
              else
                WayPoint.Free;
            end;
            8 : ReadExtension(NewTag, NewHeader, NewContent, Strict, CDataValues,
                          CodePage, Error_Msg, Warning_Msg, Succes); // extensions
            else
              Error_handling(Strict, Error_Msg, Warning_Msg, Succes,
                                  MsgUnkownElement,[NewTag, Tag]);
          end;
        end;
      end;
    until (not succes) or empty;
  end;
end;

constructor TGPXRoute.Create;
begin
  inherited;
  FWayPoints_List := TObjectList.Create;
  fDefNumber:=false;
end;

procedure TGPXRoute.DeleteWayPoint(const Index: integer);
begin
  assert((Index>=0) and (Index<FWayPoints_List.Count),'Index out of limits in TGPXRoute.DeleteWayPoints');
//  assert(Index in [0..FWayPoints_List.Count-1],'Index out of limits in TGPXRoute.DeleteWayPoints');
  FWayPoints_List.Delete(Index);
end;

destructor TGPXRoute.Destroy;
begin
  FWayPoints_List.Free;
  inherited;
end;

procedure TGPXRoute.GenerateInside(var content: TStringList;
  indent: string);
var
  I : integer;
begin
  inherited;
  if DefNumber then
    content.Add(Format('<number>%d</number>',[FNumber]));
  for I:=0 to NbWayPoints-1 do
    TGPXWayPoint(FWayPoints_List[I]).Generate(content,indent);
end;

function TGPXRoute.GetNbWayPoints: integer;
begin
  result:=FWayPoints_List.Count;
end;

function TGPXRoute.GetWayPoints(Index: Integer): TGPXWayPoint;
begin
  assert((Index>=0) and (Index<FWayPoints_List.Count),'Index out of limits in TGPXRoute.GetWayPoints');
//  assert(Index in [0..FWayPoints_List.Count-1],'Index out of limits in TGPXRoute.GetWayPoints');
  result:=FWayPoints_List[Index] as TGPXWayPoint;
end;

procedure TGPXRoute.InvalidateNumber;
begin
  fDefNumber:=false;
end;

procedure TGPXRoute.SetNumber(const Value: integer);
begin
  assert(Value>=0,'Negative value for <number> in TGPXRoute.SetNumber');
  FNumber := Value;
  FDefNumber:=true;
end;

{ TGPXTrack }

procedure TGPXTrack.AddTrackSeg(const Value: TGPXTrackSegment);
begin
  FTrackSegs_List.Add(Value);
end;

procedure TGPXTrack.Analyze_XML(var header, content: string;
  const Strict: boolean; CDataValues : array of string; const CodePage : word;
  var Error_Msg, Warning_Msg: string;
  var Succes: boolean);
var
  NewTag, NewHeader, NewContent : string;
  Empty : boolean;
  TrackSegment : TGPXTrackSegment;
begin
  if Header<>'' then
    Error_handling(Strict, Error_Msg, Warning_Msg, Succes,
                        MsgUnwantedElementHeader, [Tag,Header]);
  // content analyze
  if Succes then
  begin
    repeat
      GetFirstElement(content, NewTag, NewHeader, NewContent, Empty, Succes);
      if not Succes then
        Error_Msg := Format(MsgInvalidContent,[Tag])
      else
      begin
        if not Empty then
        begin
          Case StringIndex(NewTag, ['name', 'desc', 'cmt', 'src', 'link',
                                  'type', 'number', 'trkseg', 'extensions']) of
            0 : ReadName(NewTag, NewHeader, NewContent, Strict, CDataValues,
                                CodePage, Error_Msg, Warning_Msg, Succes); // name
            1 : ReadDescription(NewTag, NewHeader, NewContent, Strict, CDataValues,
                                CodePage, Error_Msg, Warning_Msg, Succes); // desc
            2 : ReadComment(NewTag, NewHeader, NewContent, Strict, CDataValues,
                                CodePage, Error_Msg, Warning_Msg, Succes); // cmt
            3 : ReadSource(NewTag, NewHeader, NewContent, Strict, CDataValues,
                                CodePage, Error_Msg, Warning_Msg, Succes); // src
            4 : ReadLink(NewTag, NewHeader, NewContent, Strict, CDataValues, CodePage, Error_Msg, Warning_Msg, Succes); // link
            5 : ReadCategory(NewTag, NewHeader, NewContent, Strict, CDataValues,
                                CodePage, Error_Msg, Warning_Msg, Succes); // type
            6 : begin // number
              ReplaceCData(NewContent, CDataValues);
              if DefNumber then // duplicate
                Error_handling(Strict, Error_Msg, Warning_Msg, Succes,
                                  MsgDuplicateElement,[NewTag, Tag]);
              if Succes then
              begin
                try
                  Number:=StrToInt(NewContent);
                except
                  Succes:=false;
                  Error_Msg:=Format(MsgInvalidValue,[NewTag, NewContent]);
                end;
              end;
            end;
            7 : begin // trkseg
              TrackSegment := TGPXTrackSegment.Create(NewTag);
              TrackSegment.Analyze_XML(Newheader, Newcontent, Strict, CDataValues, CodePage,
                                        Error_Msg, Warning_Msg, Succes);
              if Succes then
                AddTrackSeg(TrackSegment)
              else
                TrackSegment.Free;
            end;
            8 : ReadExtension(NewTag, NewHeader, NewContent, Strict, CDataValues,
                                CodePage, Error_Msg, Warning_Msg, Succes); // extensions
            else
              Error_handling(Strict, Error_Msg, Warning_Msg, Succes,
                                  MsgUnkownElement,[NewTag, Tag]);
          end;
        end;
      end;
    until (not succes) or empty;
  end;
end;

constructor TGPXTrack.Create;
begin
  inherited;
  FTrackSegs_List := TObjectList.Create;
end;

procedure TGPXTrack.DeleteTrackSeg(const Index: integer);
begin
  assert((Index>=0) and (Index<FTrackSegs_List.Count),'Index out of limits in TGPXTrack.DeleteTrackSegs');
//  assert(Index in [0..FTrackSegs_List.Count-1],'Index out of limits in TGPXTrack.DeleteTrackSegs');
  FTrackSegs_List.Delete(Index);
end;

destructor TGPXTrack.Destroy;
begin
  FTrackSegs_List.Free;
  inherited;
end;

procedure TGPXTrack.GenerateInside(var content: TStringList;
  indent: string);
var
  I : integer;
begin
  inherited;
  if DefNumber then
    content.Add(Format('<number>%d</number>',[FNumber]));
  for I:=0 to NbTrackSegs-1 do
    TGPXTrackSegment(FTrackSegs_List[I]).Generate(content,indent);
end;

function TGPXTrack.GetNbTrackSegs: integer;
begin
  result:=FTrackSegs_List.Count;
end;

function TGPXTrack.GetTrackSegs(Index: Integer): TGPXTrackSegment;
begin
  assert((Index>=0) and (Index<FTrackSegs_List.Count),'Index out of limits in TGPXTrack.GetTrackSegs');
//  assert(Index in [0..FTrackSegs_List.Count-1],'Index out of limits in TGPXTrack.GetTrackSegs');
  result:=FTrackSegs_List[Index] as TGPXTrackSegment;
end;

procedure TGPXTrack.InvalidateNumber;
begin
  FDefNumber := false;
end;

procedure TGPXTrack.SetNumber(const Value: integer);
begin
  assert(Value>=0,'Negative value for <number> in TGPXTrack.SetNumber');
  FNumber := Value;
  FDefNumber:=true;
end;

{ TGPXData }

procedure TGPXData.AddRoute(const Value: TGPXRoute);
begin
  FRoute_List.Add(Value);
end;

procedure TGPXData.AddTrack(const Value: TGPXTrack);
begin
  FTrack_List.Add(Value);
end;

procedure TGPXData.AddWayPoint(const Value: TGPXWaypoint);
begin
  FWayPoints_List.Add(Value);
end;

constructor TGPXData.Create(const Creator_Value: string);
begin
  inherited Create('gpx');
  Creator:=Creator_Value;
  FWayPoints_List:=TObjectList.Create;
  FRoute_List:=TObjectList.Create;
  FTrack_List:=TObjectList.Create;
end;

destructor TGPXData.Destroy;
begin
  FWayPoints_List.Free;
  FRoute_List.Free;
  FTrack_List.Free;
  DeleteMetaData;
  inherited;
end;

procedure TGPXData.Generate(var content: TStringList);
begin
  assert(Creator<>'','No Creator defined in TGPXData.Generate');
  content.Add('<?xml version="1.0" encoding="UTF-8"?>');
  inherited Generate(content,'');
end;

function TGPXData.GetRoutes(Index: Integer): TGPXRoute;
begin
  assert((Index>=0) and (Index<FRoute_List.Count),'Index out of limits in TGPXData.GetRoutes');
//  assert(Index in [0..FRoute_List.Count-1],'Index out of limits in TGPXData.GetRoutes');
  result:=FRoute_List[Index] as TGPXRoute;
end;

function TGPXData.GetTracks(Index: Integer): TGPXTrack;
begin
  assert((Index>=0) and (Index<FTrack_List.Count),'Index out of limits in TGPXData.GetTracks');
//  assert(Index in [0..FTrack_List.Count-1],'Index out of limits in TGPXData.GetTracks');
  result:=FTrack_List[Index] as TGPXTrack;
end;

function TGPXData.GetWayPoints(Index: Integer): TGPXWaypoint;
begin
  assert((Index>=0) and (Index<FWayPoints_List.Count),'Index out of limits in TGPXData.GetWayPoints');
//  assert(Index in [0..FWayPoints_List.Count-1],'Index out of limits in TGPXData.GetWayPoints');
  result:=FWayPoints_List[Index] as TGPXWayPoint;
end;

procedure TGPXData.DeleteMetadata;
begin
  if FMetadata<>nil then
  begin
    FMetaData.Free;
    FMetaData:=nil;
  end;
end;

procedure TGPXData.SetCreator(const Value: string);
begin
  FCreator := Value;
end;

procedure TGPXData.SetMetadata(const Value: TGPXMetaData);
begin
  FMetadata := Value;
end;

procedure TGPXData.DeleteRoute(const Index: integer);
begin
  assert((Index>=0) and (Index<FRoute_List.Count),'Index out of limits in TGPXData.DeleteRoute');
//  assert(Index in [0..FRoute_List.Count-1],'Index out of limits in TGPXData.DeleteRoute');
  FRoute_List.Delete(Index);
end;

procedure TGPXData.DeleteTrack(const Index: integer);
begin
  assert((Index>=0) and (Index<FTrack_List.Count),'Index out of limits in TGPXData.DeleteTrack');
//  assert(Index in [0..FTrack_List.Count-1],'Index out of limits in TGPXData.DeleteTrack');
  FTrack_List.Delete(Index);
end;

procedure TGPXData.DeleteWayPoint(const Index: integer);
begin
  assert((Index>=0) and (Index<FWayPoints_List.Count),'Index out of limits in TGPXData.DeleteWayPoint');
//  assert(Index in [0..FWayPoints_List.Count-1],'Index out of limits in TGPXData.DeleteWayPoint');
  FWayPoints_List.Delete(Index);
end;

function TGPXData.GetNbRoutes: integer;
begin
 result:=FRoute_List.Count;
end;

function TGPXData.GetNbTracks: integer;
begin
 result:=FTrack_List.Count;
end;

function TGPXData.GetNbWaypoints: integer;
begin
 result:=FWayPoints_List.Count;
end;

procedure TGPXData.SaveToFile(const FileName: string);
var
  F : TextFile;
  Content : TStringList;
  I : integer;
begin
  assignFile(F, FileName);
  rewrite(F);
  try
    Content := TStringList.Create;
    Generate(Content);
    for I:=0 to Content.Count-1 do
      writeln(F,Content[I]);
  finally
    CloseFile(F);
  end;
end;

procedure TGPXData.LoadFromFile(const FileName: string; const Strict : boolean;
                                var Error_Msg, Warning_Msg : string;
                                var Succes : boolean);
// Strict indicate if it should stop on unkown parts of the file
// Error_Msg (resp. Warning_Msg) return the list of errors (resp. warning)
// Succes is set at True if not error found
var
  Nb_Ligne, Nb_Char : integer;
  F : TextFile;
  G : File;
  XML_Content, S : string;
  TheTag, Header, Content : string;
  xName, xValue : string;
  Empty, EncodingFound : boolean;
  CDataValue : array of string;
  CodePage : word;

  procedure FindCData(var S : string; var Succes : boolean);
  var
    Current, Start, Stop : integer;
  begin
    Succes:=true;
    Current:=1;
    while Succes and (PosAfter('<![CDATA[',S,Current)<>0) do
    begin
      Start:=PosAfter('<![CDATA[',S,Current);
      Stop:=PosAfter(']]>',S,Start);
      if Stop=0 then
        Succes:=false
      else
      begin
        setlength(CDataValue,high(CDataValue)+2);
        CDataValue[high(CDataValue)]:=copy(S,Start+8,Stop-Start-9);
        delete(S,Start+8,Stop-Start-9);
        insert(format('%.*d',[Cdata_digit, high(CDataValue)]), S, Start+8);
        if high(CDataValue)>IntPower(10.0, CData_digit)-1 then
          raise ERangeError.Create('Too much CDATA sequences');
        Current:=Start+11;
      end;
    end;
  end;


  procedure One;
  var
    Nb_Ligne, Nb_Char : integer;
    F : TextFile;
    G : File;
  begin
  // loading the GPX file in a single string
  try
    assignFile(G, FileName);
    reset(G,1);
    try
      setlength(XML_Content,filesize(G));
    finally
      CloseFile(G);
    end;


    assignFile(F, FileName);
    reset(F);
    try

      Nb_Ligne := 0;
      Nb_Char :=1;
      while not Eof(F) do
      begin
        inc(Nb_Ligne);
        readln(F,S);
        move(S[1], XML_Content[Nb_Char], length(S));
        inc(Nb_Char, length(S));
      end;
      XML_Content := copy(XML_Content,1, length(XML_Content)-(Nb_Ligne-1)*2);  // removing unused char due to CR+LF at each end of line in the file
    finally
      CloseFile(F);
    end;
  except
    Succes:=false;
    Error_Msg:=MsgCantRead;
  end;
  end;

  procedure Two;
  begin
  if Succes then
  begin
    RemoveXMLComments(XML_Content, Succes);
    if not Succes then
      Error_Msg:=MsgInvalidXMLComment;
  end;
  end;

  procedure Three;
  begin
  if Succes then
  begin
    FindCData(XML_Content, Succes);
    if not Succes then
      Error_Msg:=MsgInvalidCData;
  end;
  end;

  procedure Four;
  begin
  if Succes then
  begin
  // removing Tab, CR and LF
  // this could be a bad thing for multiline elements versus xml definition
  // but it should not be a problem with GPX
    while pos(#9,XML_Content)<>0 do
      delete(XML_Content,pos(#9,XML_Content),1);
    while pos(#10,XML_Content)<>0 do
      delete(XML_Content,pos(#10,XML_Content),1);
    while pos(#13,XML_Content)<>0 do
      delete(XML_Content,pos(#13,XML_Content),1);
  end;
  end;

  procedure Six;
  begin
  // analyzing xml header for encoding
  if Succes then
  begin
    if pos('<?xml',XML_Content)=0 then
    begin
      Succes:=false;
      Error_Msg := MsgXMLHeaderNotFound;
    end
    else
    begin
      delete(XML_Content, 1, pos('<?xml',XML_Content)+5);
      if pos('?>',XML_Content)=0 then
      begin
        Succes:=false;
        Error_Msg := MsgInvalidXMLHeader;
      end
      else
      begin
        EncodingFound:=false;
        Content:=copy(XML_Content, 1, pos('?>',XML_Content)-1);
        delete(XML_Content, 1, pos('?>',XML_Content)+1);
        repeat
          GetFirstValue(Content, xName, xValue, Empty, Succes);
          if not Succes then
            Error_Msg:= MsgInvalidXMLHeader
          else
          begin
            if not Empty then
            begin
              if Succes and (xName='encoding') then
              begin
                if EncodingFound then // duplicate
                  Error_handling(Strict, Error_Msg, Warning_Msg, Succes,
                                 MsgDuplicateField,[xName,'xml']);
                CodePage:=ALGetCodePageFromName(xValue);
                EncodingFound:=true;
              end;
            end;
          end;
        until not Succes or Empty;
        if not EncodingFound then
          Warning_Msg:=Warning_Msg+#13#10+Format(MsgMissingField,['encoding', 'xml']);
      end;
    end;
  end;
  end;

  procedure Seven;
  begin
  // removing other header(s) <? ... ?>
  while Succes and (pos('<?',XML_Content)<>0) do
  begin
    delete(XML_Content, 1, pos('<?',XML_Content)+1);
    if pos('?>',XML_Content)=0 then
    begin
      Succes:=false;
      Error_Msg := MsgInvalidXMLHeader;
    end
    else
      delete(XML_Content, 1, pos('?>',XML_Content)+1);
  end;
  end;

  procedure Eight;
  begin
  if Succes then
  begin
    GetFirstElement(XML_Content, TheTag, Header, Content, Empty, Succes);
    if not Succes then
      Error_Msg:=MsgInvalidMainElement
    else
    begin
      if Empty then
      begin
        Succes:=false;
        Error_Msg := MsgMissingMainElement;
      end
      else
      begin
        if TheTag<>'gpx' then
        begin
          Succes:=false;
          Error_Msg := MsgWrongMainElement;
        end
        else
        begin
          Analyze_XML(header, content, Strict, CDataValue, CodePage, Error_Msg, Warning_Msg, Succes);
          if Succes and Strict then
          begin // checking for others <elements> after <gpx>
            GetFirstElement(XML_Content, TheTag, Header, Content, Empty, Succes);
            if not Succes or not Empty then
            begin
              Succes:=false;
              Error_Msg:=MsgAfterEnd;
            end;
          end;
        end;
      end;
    end;
  end;
  end;
{
begin
  Succes:=true;
  Error_Msg:='';
  Warning_Msg:='';
  CodePage:=0;
  One;
  Two;
  Three;
  Four;
  Six;
  Seven;
  Eight;
 }

begin
  Succes:=true;
  Error_Msg:='';
  Warning_Msg:='';
  CodePage:=0;

  // loading the GPX file in a single string
  try
    assignFile(G, FileName);
    reset(G,1);
    try
      setlength(XML_Content,filesize(G));
    finally
      CloseFile(G);
    end;


    assignFile(F, FileName);
    reset(F);
    try

(*  First version, very slow due to memory re-allocation at each line added
      while not Eof(F) do
      begin
        readln(F,S);
        XML_Content:=XML_Content+S;
      end;
*)


      Nb_Ligne := 0;
      Nb_Char :=1;
      while not Eof(F) do
      begin
        inc(Nb_Ligne);
        readln(F,S);
        move(S[1], XML_Content[Nb_Char], length(S));
        inc(Nb_Char, length(S));
      end;
      XML_Content := copy(XML_Content,1, length(XML_Content)-(Nb_Ligne-2)*2);  // removing unused char due to CR+LF at each end of line in the file
    finally
      CloseFile(F);
    end;
  except
    Succes:=false;
    Error_Msg:=MsgCantRead;
  end;

  if Succes then
  begin
    FindCData(XML_Content, Succes);
    if not Succes then
      Error_Msg:=MsgInvalidCData;
  end;

  if Succes then
  begin
  // removing Tab, CR and LF
  // this could be a bad thing for multiline elements versus xml definition
  // but it should not be a problem with GPX
    while pos(#9,XML_Content)<>0 do
      delete(XML_Content,pos(#9,XML_Content),1);
    while pos(#10,XML_Content)<>0 do
      delete(XML_Content,pos(#10,XML_Content),1);
    while pos(#13,XML_Content)<>0 do
      delete(XML_Content,pos(#13,XML_Content),1);
  end;

  // analyzing xml header for encoding
  if Succes then
  begin
    if pos('<?xml',XML_Content)=0 then
    begin
      Succes:=false;
      Error_Msg := MsgXMLHeaderNotFound;
    end
    else
    begin
      delete(XML_Content, 1, pos('<?xml',XML_Content)+5);
      if pos('?>',XML_Content)=0 then
      begin
        Succes:=false;
        Error_Msg := MsgInvalidXMLHeader;
      end
      else
      begin
        EncodingFound:=false;
        Content:=copy(XML_Content, 1, pos('?>',XML_Content)-1);
        delete(XML_Content, 1, pos('?>',XML_Content)+1);
        repeat
          GetFirstValue(Content, xName, xValue, Empty, Succes);
          if not Succes then
            Error_Msg:= MsgInvalidXMLHeader
          else
          begin
            if not Empty then
            begin
              if Succes and (xName='encoding') then
              begin
                if EncodingFound then // duplicate
                  Error_handling(Strict, Error_Msg, Warning_Msg, Succes,
                                 MsgDuplicateField,[xName,'xml']);
                CodePage:=ALGetCodePageFromName(xValue);
                EncodingFound:=true;
              end;
            end;
          end;
        until not Succes or Empty;
        if not EncodingFound then
          Warning_Msg:=Warning_Msg+#13#10+Format(MsgMissingField,['encoding', 'xml']);
      end;
    end;
  end;

  // removing other header(s) <? ... ?>
  while Succes and (pos('<?',XML_Content)<>0) do
  begin
    delete(XML_Content, 1, pos('<?',XML_Content)+1);
    if pos('?>',XML_Content)=0 then
    begin
      Succes:=false;
      Error_Msg := MsgInvalidXMLHeader;
    end
    else
      delete(XML_Content, 1, pos('?>',XML_Content)+1);
  end;

  if Succes then
  begin
    GetFirstElement(XML_Content, TheTag, Header, Content, Empty, Succes);
    if not Succes then
      Error_Msg:=MsgInvalidMainElement
    else
    begin
      if Empty then
      begin
        Succes:=false;
        Error_Msg := MsgMissingMainElement;
      end
      else
      begin
        if TheTag<>'gpx' then
        begin
          Succes:=false;
          Error_Msg := MsgWrongMainElement;
        end
        else
        begin
          Analyze_XML(header, content, Strict, CDataValue, CodePage, Error_Msg, Warning_Msg, Succes);
          if Succes and Strict then
          begin // checking for others <elements> after <gpx>
            GetFirstElement(XML_Content, TheTag, Header, Content, Empty, Succes);
            if not Succes or not Empty then
            begin
              Succes:=false;
              Error_Msg:=MsgAfterEnd;
            end;
          end;
        end;
      end;
    end;
  end;

end;


procedure TGPXData.Analyze_XML(var header, content: string;
  const Strict: boolean; CDataValues : array of string; const CodePage : word;
  var Error_Msg, Warning_Msg: string;
  var Succes: boolean);
var
  Empty, Creator_found, Version_found : boolean;
  Name, Value : string;
  NewTag, NewHeader, NewContent : string;
  WayPoint : TGPXWayPoint;
  Route : TGPXRoute;
  Track : TGPXTrack;
begin
  Succes:=true;
  Creator_found :=false; Version_found:=false;
  // header analyze
  repeat
    GetFirstValue(header, Name, Value, Empty, Succes);
    ReplaceCData(Value, CDataValues);
    if not Succes then
      Error_Msg:=Format(MsgInvalidElementHeader,[Tag])
    else
    begin
      if not Empty then
      begin
        Case StringIndex(Name, ['creator', 'version', 'xmlns', 'xmlns:xsi', 'xsi:schemaLocation']) of
          0 : begin // creator, not stored
            if Creator_Found then // duplicate
              Error_handling(Strict, Error_Msg, Warning_Msg, Succes,
                               MsgDuplicateField,[Name,Tag]);
            Creator_found:=true;
          end;
          1 : begin // version
            if Version_Found then // duplicate
              Error_handling(Strict, Error_Msg, Warning_Msg, Succes,
                               MsgDuplicateField,[Name,Tag]);
            if Succes then
            begin
              if Value<>'1.1' then
                Error_handling(Strict, Error_Msg, Warning_Msg, Succes,
                               MsgWrongVersion,[Value]);
              Version_Found:=true;
            end;
          end;
          2..4 : begin // xmlns xmlns:xsi xsi, not analyzed
          end;
          else
            Error_handling(Strict, Error_Msg, Warning_Msg, Succes,
                               MsgUnknownField,[Name, Tag]);
        end;
      end;
    end;
  until (not Succes) or empty;

  if Succes and not Version_Found then
    Error_handling(Strict, Error_Msg, Warning_Msg, Succes,
                               MsgMissingField,['version',Tag]);

  if Succes and not Creator_Found then
    Error_handling(Strict, Error_Msg, Warning_Msg, Succes,
                               MsgMissingField,['creator',Tag]);

  // content analyze
  if Succes then
  begin
    repeat
      GetFirstElement(content, NewTag, NewHeader, NewContent, Empty, Succes);
      if not Succes then
        Error_Msg := Format(MsgInvalidContent,[Tag])
      else
      begin
        if not Empty then
        begin
          Case StringIndex(NewTag, ['metadata', 'wpt', 'rte', 'trk', 'extensions']) of
            0 : begin // metadata
              if Metadata<>nil then
              begin
                Error_handling(Strict, Error_Msg, Warning_Msg, Succes,
                               MsgDuplicateElement,[NewTag,Tag]);
                if Succes then
                  DeleteMetadata;
              end;
              if Succes then
              begin
                Metadata:=TGPXMetadata.Create(NewTag);
                Metadata.Analyze_XML(Newheader, Newcontent, Strict, CDataValues, CodePage,
                                        Error_Msg, Warning_Msg, Succes);
              end;
            end;
            1 : begin // wpt
              WayPoint := TGPXWayPoint.Create(NewTag);
              WayPoint.Analyze_XML(Newheader, Newcontent, Strict, CDataValues, CodePage, 
                                        Error_Msg, Warning_Msg, Succes);
              if Succes then
                AddWayPoint(WayPoint)
              else
                WayPoint.Free;
            end;
            2 : begin // rte
              Route := TGPXRoute.Create(NewTag);
              Route.Analyze_XML(Newheader, Newcontent, Strict, CDataValues, CodePage, 
                                        Error_Msg, Warning_Msg, Succes);
              if Succes then
                AddRoute(Route)
              else
                Route.Free;
            end;
            3 : begin // trk
              Track := TGPXTrack.Create(NewTag);
              Track.Analyze_XML(Newheader, Newcontent, Strict, CDataValues, CodePage, 
                                        Error_Msg, Warning_Msg, Succes);
              if Succes then
                AddTrack(Track)
              else
                Track.Free;
            end;
            4 : begin // extension
              ReadExtension (NewTag, Newheader, Newcontent, Strict, CDataValues,
                                        CodePage, Error_Msg, Warning_Msg, Succes);
            end;
            else
              Error_handling(Strict, Error_Msg, Warning_Msg, Succes,
                               MsgUnkownElement,[NewTag, Tag]);
          end;
        end;
      end;
    until (not succes) or empty;
  end;
end;

function TGPXData.GenerateHeader: string;
begin
  result:=format(' version="1.1" creator=%s',[QuoteXMLString(Creator)])+#13#10+
        ' xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance"'+#13#10+
        ' xmlns="http://www.topografix.com/GPX/1/1"'+#13#10+
        ' xsi:schemaLocation="http://www.topografix.com/GPX/1/1 http://www.topografix.com/GPX/1/1/gpx.xsd"';
end;

procedure TGPXData.GenerateInside(var content: TStringList;
  indent: string);
var
  I : integer;
begin
  inherited;

  if MetaData<>nil then
    Metadata.Generate(content, one_indent);
  for I:=0 to NbWaypoints-1 do
    WayPoints[I].Generate(content, one_indent);
  for I:=0 to NbRoutes-1 do
    Routes[I].Generate(content, one_indent);
  for I:=0 to NbTracks-1 do
    Tracks[I].Generate(content, one_indent);
end;

procedure TGPXData.Clear;
begin
  FWayPoints_List.Clear;
  FRoute_List.Clear;
  FTrack_List.Clear;
  DeleteMetaData;
end;

{ TGPXTrackSegment }

procedure TGPXTrackSegment.AddWayPoint(const Value: TGPXWayPoint);
begin
  FWayPoints_List.Add(Value);
end;

procedure TGPXTrackSegment.Analyze_XML(var header, content: string;
  const Strict: boolean; CDataValues : array of string; const CodePage : word;
  var Error_Msg, Warning_Msg: string;
  var Succes: boolean);
var
  NewTag, NewHeader, NewContent : string;
  Empty : boolean;
  WayPoint : TGPXWayPoint;
begin
  if Header<>'' then
    Error_handling(Strict, Error_Msg, Warning_Msg, Succes,
                        MsgUnwantedElementHeader, [Tag,Header]);
  // content analyze
  if Succes then
  begin
    repeat
      GetFirstElement(content, NewTag, NewHeader, NewContent, Empty, Succes);
      if not Succes then
        Error_Msg := Format(MsgInvalidContent,[Tag])
      else
      begin
        if not Empty then
        begin
          Case StringIndex(NewTag, ['trkpt', 'extensions']) of
            0 : begin // trkpt
              WayPoint := TGPXWayPoint.Create(NewTag);
              WayPoint.Analyze_XML(Newheader, Newcontent, Strict, CDataValues, CodePage,
                                        Error_Msg, Warning_Msg, Succes);
              if Succes then
                AddWayPoint(WayPoint)
              else
                WayPoint.Free;
            end;
            1 : ReadExtension(NewTag, NewHeader, NewContent, Strict, CDataValues,
                                CodePage, Error_Msg, Warning_Msg, Succes); // extensions
            else
              Error_handling(Strict, Error_Msg, Warning_Msg, Succes,
                                  MsgUnkownElement,[NewTag, Tag]);
          end;
        end;
      end;
    until (not succes) or empty;
  end;
end;

constructor TGPXTrackSegment.Create;
begin
  inherited;
  FWayPoints_List := TObjectList.Create;
end;

procedure TGPXTrackSegment.DeleteWayPoint(const Index: integer);
begin
  assert((Index>=0) and (Index<FWayPoints_List.Count),'Index out of limits in TGPXTrackSegment.DeleteWayPoint');
//  assert(Index in [0..FWayPoints_List.Count-1],'Index out of limits in TGPXTrackSegment.DeleteWayPoint');
  FWayPoints_List.Delete(Index);
end;

destructor TGPXTrackSegment.Destroy;
begin
  FWayPoints_List.Free;
end;

procedure TGPXTrackSegment.GenerateInside(var content: TStringList;
  indent: string);
var
  I : integer;
begin
  inherited;
  for I:=0 to NbWayPoints-1 do
    TGPXWayPoint(FWayPoints_List[I]).Generate(content,indent);
end;


function TGPXTrackSegment.GetNbWayPoints: integer;
begin
  result:=FWayPoints_List.Count;
end;

function TGPXTrackSegment.GetWayPoints(Index: Integer): TGPXWayPoint;
begin
  assert((Index>=0) and (Index<FWayPoints_List.Count),'Index out of limits in TGPXTrackSegment.GetWayPoints');
  result:= FWayPoints_List[Index] as TGPXWayPoint;
end;


{ TGPXPerson }

destructor TGPXPerson.Destroy;
begin
  DeleteLink;
  inherited;
end;

procedure TGPXPerson.GenerateInside(var content: TStringList;
  indent: string);
begin
  inherited;
  if Name<>'' then
    content.Add(Format('%s<name>%s</name>',[indent,ProtectXML(Name)]));
  if Email_id<>'' then
    content.Add(Format('%s<email id="%s" domain="%s" />',[indent, Email_id, Email_domain]));
  if FLink<>nil then
    FLink.Generate(content, indent+one_indent);

end;

procedure TGPXPerson.DeleteLink;
begin
  if FLink<>nil then
  begin
    FLink.Free;
    FLink:=nil;
  end;
end;

procedure TGPXPerson.SetEmail_domain(const Value: string);
begin
  FEmail_domain := Value;
end;

procedure TGPXPerson.SetEmail_id(const Value: string);
begin
  FEmail_id := Value;
end;

procedure TGPXPerson.SetLink(const Value: TGPXLink);
begin
  FLink := Value;
end;

procedure TGPXPerson.SetName(const Value: string);
begin
  FName := Value;
end;

procedure TGPXPerson.Analyze_XML(var header, content: string;
  const Strict: boolean; CDataValues : array of string; const CodePage : word;
  var Error_Msg, Warning_Msg: string;
  var Succes: boolean);
var
  NewTag, NewHeader, NewContent : string;
  Empty : boolean;
  eName, eValue : string;
  eEmpty, id_found, domain_found : boolean;
begin
  if Header<>'' then
    Error_handling(Strict, Error_Msg, Warning_Msg, Succes,
                        MsgUnwantedElementHeader, [Tag,Header]);
  // content analyze
  if Succes then
  begin
    repeat
      GetFirstElement(content, NewTag, NewHeader, NewContent, Empty, Succes);
      if not Succes then
        Error_Msg := Format(MsgInvalidContent,[Tag])
      else
      begin
        if not Empty then
        begin
          Case StringIndex(NewTag, ['name', 'email', 'link']) of
            0 : begin // name
              ReplaceCData(NewContent, CDataValues);
              if Name<>'' then
                Error_handling(Strict, Error_Msg, Warning_Msg, Succes,
                                 MsgDuplicateElement,[NewTag,Tag]);
              if Succes then
                Name:=ALUTF8Encode(NewContent, CodePage);
            end;
            1 : begin // email
              if Email_id<>'' then
                Error_handling(Strict, Error_Msg, Warning_Msg, Succes,
                                 MsgDuplicateElement,[NewTag,Tag]);
              if Succes and (NewContent<>'') then
                Error_handling(Strict, Error_Msg, Warning_Msg, Succes,
                                  MsgUnwantedElementContent,[NewTag,NewContent]);
              if Succes then
              begin
                id_found:=false;
                domain_found:=false;
                repeat
                  GetFirstValue(NewHeader, eName, eValue, eEmpty, Succes);
                  ReplaceCData(eValue, CDataValues);
                  if not Succes then
                    Error_Msg:=Format(MsgInvalidElementHeader,[NewTag]);
                  if Succes and not eEmpty then
                  begin
                    case StringIndex(eName, ['id', 'domain']) of
                      0 : begin // id
                        if id_found then
                          Error_handling(Strict, Error_Msg, Warning_Msg, Succes,
                                  MsgDuplicateField,[eName, NewTag]);
                        if Succes then
                        begin
                          Email_id:=ALUTF8Encode(eValue, CodePage);
                          id_found:=true;
                        end;
                      end;
                      1 : begin // domain
                        if domain_found then
                          Error_handling(Strict, Error_Msg, Warning_Msg, Succes,
                                  MsgDuplicateField,[eName, NewTag]);
                        if Succes then
                        begin
                          Email_domain:=ALUTF8Encode(eValue, CodePage);
                          domain_found:=true;
                        end;
                      end;
                      else
                        Error_handling(Strict, Error_Msg, Warning_Msg, Succes,
                                  MsgUnknownField,[eName, NewTag]);
                    end;
                  end;
                until (not Succes) or eEmpty;
                if Succes and not id_found then
                  Error_handling(Strict, Error_Msg, Warning_Msg, Succes,
                                  MsgMissingField,['id', NewTag]);
                if Succes and not domain_found then
                  Error_handling(Strict, Error_Msg, Warning_Msg, Succes,
                                  MsgMissingField,['domain', NewTag]);
              end;
            end;
            2 : begin  // link
              if Link<>nil then
              begin
                Error_handling(Strict, Error_Msg, Warning_Msg, Succes,
                               MsgDuplicateElement,[NewTag,Tag]);
                if Succes then
                  DeleteLink;
              end;
              if Succes then
              begin
                Link:=TGPXLink.Create(NewTag);
                Link.Analyze_XML(Newheader, Newcontent, Strict, CDataValues, CodePage, 
                                        Error_Msg, Warning_Msg, Succes);
              end;
            end;
            else
              Error_handling(Strict, Error_Msg, Warning_Msg, Succes,
                                  MsgUnkownElement,[NewTag, Tag]);
          end;
        end;
      end;
    until (not succes) or empty;
  end;
end;

{ TGpxBasic }

constructor TGpxBasic.Create(const Tag_Value : string);
begin
  Tag := Tag_Value;
end;

procedure TGpxBasic.Generate(var content: TStringList; indent: string);
begin
  content.Add(Format('%s<%s%s>',[indent,Tag, GenerateHeader]));

  GenerateInside(content,indent+one_indent);

  content.Add(Format('%s</%s>',[indent,Tag]));
end;

function TGpxBasic.GenerateHeader: string;
begin
  result:='';
end;

procedure TGpxBasic.SetTag(const Value: string);
begin
  FTag := Value;
end;

{ TGPXCopyright }

procedure TGPXCopyright.Analyze_XML(var header, content: string;
  const Strict: boolean; CDataValues : array of string; const CodePage : word;
  var Error_Msg, Warning_Msg: string;
  var Succes: boolean);
var
  Author_Found, Empty : boolean;
  Name, Value : string;
  NewTag, NewHeader, NewContent : string;
begin
  Succes:=true;
  Author_Found:=false;
  // header analyze
  repeat
    GetFirstValue(header, Name, Value, Empty, Succes);
    ReplaceCData(Value, CDataValues);
    if not Succes then
      Error_Msg:=Format(MsgInvalidElementHeader,[Tag])
    else
    begin
      if not Empty then
      begin
        if Name='author' then
        begin
          if Author_Found then // duplicate
            Error_handling(Strict, Error_Msg, Warning_Msg, Succes,
                           MsgDuplicateField,[Name,Tag]);
          if Succes then
          begin
            Author:=ALUTF8Encode(Value, CodePage);
            Author_Found:=true;
          end;
        end
        else
          Error_handling(Strict, Error_Msg, Warning_Msg, Succes,
                           MsgUnknownField,[Name,Tag]);
      end;
    end;
  until (not Succes) or empty;

  if Succes and not Author_Found then
  begin
    Succes:=false;
    Error_Msg:=Format(MsgMissingField,['author',Tag]);
  end;

  // content analyze
  if Succes then
  begin
    repeat
      GetFirstElement(content, NewTag, NewHeader, NewContent, Empty, Succes);
      if not Succes then
        Error_Msg := Format(MsgInvalidContent,[Tag])
      else
      begin
        if not Empty then
        begin
          Case StringIndex(NewTag, ['year', 'license']) of
            0 : begin // year
              ReplaceCData(NewContent, CDataValues);
              if DefYear then
                Error_handling(Strict, Error_Msg, Warning_Msg, Succes,
                               MsgDuplicateElement,[NewTag,Tag]);
              if Succes then
              begin
                try
                  Year:=StrToInt(NewContent);
                except
                  Succes:=false;
                  Error_Msg:=Format(MsgInvalidValue,[NewTag, NewContent]);
                end;
              end;
            end;
            1 : begin // license
              ReplaceCData(NewContent, CDataValues);
              if License<>'' then
                Error_handling(Strict, Error_Msg, Warning_Msg, Succes,
                               MsgDuplicateElement,[NewTag,Tag]);
              if Succes then
                License:=ALUTF8Encode(NewContent, CodePage);
            end;
            else
              Error_handling(Strict, Error_Msg, Warning_Msg, Succes,
                               MsgUnkownElement,[NewTag,Tag]);
          end;
        end;
      end;
    until (not succes) or empty;
  end;
end;

constructor TGPXCopyright.Create(const Tag_Value: string);
begin
  inherited;
  FDefYear:=false;
end;

function TGPXCopyright.GenerateHeader: string;
begin
  result:=Format(' author=%s',[QuoteXMLString(Author)]);
end;

procedure TGPXCopyright.GenerateInside(var content: TStringList;
  indent: string);
begin
  inherited;
  if DefYear then
    content.Add(Format('%s<year>%d</year>',[indent,Year]));
  if License<>'' then
    content.Add(Format('%s<license>%s</license>',[indent, ProtectXML(License)]));
end;

procedure TGPXCopyright.InvalidateYear;
begin
  FDefYear:=false;
end;

procedure TGPXCopyright.SetAuthor(const Value: string);
begin
  FAuthor := Value;
end;

procedure TGPXCopyright.SetLicense(const Value: string);
begin
  FLicense := Value;
end;

procedure TGPXCopyright.SetYear(const Value: integer);
begin
  FYear := Value;
  FDefYear:=true;
end;

{ TGPXMetaData }

constructor TGPXMetaData.Create(const Tag_Value: string);
begin
  inherited;
  FDefBounds := false;
  FDefTime := false;
end;

destructor TGPXMetaData.Destroy;
begin
  DeleteCopyright;
  DeleteAuthor;
  inherited;
end;

procedure TGPXMetaData.GenerateInside(var content: TStringList;
  indent: string);
begin
  inherited;
  if Author<>nil then
    Author.Generate(content, indent+one_indent);
  if Copyright<>nil then
    Copyright.Generate(content, indent+one_indent);
  if DefTime then
    content.Add(Format('%s<time>%.4d-%.2d-%.2dT%.2d:%.2d:%.2dZ</time>',
                       [indent, YearOf(Time), MonthOf(Time), DayOf(Time),
                       HourOf(Time), MinuteOf(Time), SecondOf(Time) ]));
  if Keywords<>'' then
    content.Add(Format('%s<keywords>%s</keywords>',[indent,ProtectXML(Keywords)]));
  if DefBounds then
    content.Add(StringReplace(Format('%s<bounds minlat="%.4f" minlon="%.4f" maxlat="%.4f" maxlon="%.4f" />',
              [indent, Bounds.LatMin, Bounds.LongMin, Bounds.LatMax, Bounds.LongMax] ),',','.',[rfReplaceAll]));
end;

procedure TGPXMetaData.InvalidateBounds;
begin
  FDefBounds:=false;
end;

procedure TGPXMetaData.InvalidateTime;
begin
  FDefTime:=false;
end;

procedure TGPXMetaData.DeleteAuthor;
begin
  if FAuthor<>nil then
  begin
    FAuthor.Free;
    FAuthor:=nil;
  end;
end;

procedure TGPXMetaData.DeleteCopyright;
begin
  if FCopyright<>nil then
  begin
    FCopyright.Free;
    FCopyright:=nil;
  end;
end;

procedure TGPXMetaData.SetAuthor(const Value: TGPXPerson);
begin
  FAuthor := Value;
end;

procedure TGPXMetaData.SetBounds(const Value: TBound);
begin
  FBounds := Value;
  FDefBounds:=true;
end;

procedure TGPXMetaData.SetCopyright(const Value: TGPXCopyright);
begin
  FCopyright := Value;
end;

procedure TGPXMetaData.SetKeywords(const Value: string);
begin
  FKeywords := Value;
end;

procedure TGPXMetaData.SetTime(const Value: TDateTime);
begin
  FTime := Value;
  FDefTime := true;
end;

procedure TGPXMetaData.Analyze_XML(var header, content: string;
  const Strict: boolean; CDataValues : array of string; const CodePage : word;
  var Error_Msg, Warning_Msg: string;
  var Succes: boolean);
var
  NewTag, NewHeader, NewContent : string;
  Empty : boolean;
  bName, bValue : string;
  bEmpty, lat_min_found, lat_max_found, lon_min_found, lon_max_found : boolean;
  Erreur : integer;
  TheTime : TDateTime;
  TheBounds : TBound;
begin
  if Header<>'' then
    Error_handling(Strict, Error_Msg, Warning_Msg, Succes,
                        MsgUnwantedElementHeader, [Tag,Header]);
  // content analyze
  if Succes then
  begin
    repeat
      GetFirstElement(content, NewTag, NewHeader, NewContent, Empty, Succes);
      if not Succes then
        Error_Msg := Format(MsgInvalidContent,[Tag])
      else
      begin
        if not Empty then
        begin
          Case StringIndex(NewTag, ['name', 'desc', 'author', 'copyright',
                         'link', 'time', 'keywords', 'bounds', 'extensions']) of
            0 : ReadName(NewTag, NewHeader, NewContent, Strict, CDataValues,
                                CodePage, Error_Msg, Warning_Msg, Succes); // name
            1 : ReadDescription(NewTag, NewHeader, NewContent, Strict, CDataValues,
                                CodePage, Error_Msg, Warning_Msg, Succes); // desc
            2 : begin // author
              if Author<>nil then
              begin
                Error_handling(Strict, Error_Msg, Warning_Msg, Succes,
                               MsgDuplicateElement,[NewTag,Tag]);
                if Succes then
                  DeleteAuthor;
              end;
              if Succes then
              begin
                Author:=TGPXPerson.Create(NewTag);
                Author.Analyze_XML(Newheader, Newcontent, Strict, CDataValues, CodePage,
                                        Error_Msg, Warning_Msg, Succes);
              end;
            end;
            3 : begin // copyright
              if Copyright<>nil then
              begin
                Error_handling(Strict, Error_Msg, Warning_Msg, Succes,
                               MsgDuplicateElement,[NewTag,Tag]);
                if Succes then
                  DeleteCopyright;
              end;
              if Succes then
              begin
                Copyright:=TGPXCopyright.Create(NewTag);
                Copyright.Analyze_XML(Newheader, Newcontent, Strict, CDataValues, CodePage,
                                        Error_Msg, Warning_Msg, Succes);
              end;
            end;
            4 : ReadLink(NewTag, NewHeader, NewContent, Strict, CDataValues, CodePage,
                                             Error_Msg, Warning_Msg, Succes); // link
            5 : begin // time
              ReplaceCData(NewContent, CDataValues);
              GetTimeISO8601(NewContent, TheTime, Succes);
              if Succes then
                Time:=TheTime
              else
                Error_Msg:=Format(MsgInvalidTime,[NewContent,Tag]);
            end;
            6 : begin // keywords
              ReplaceCData(NewContent, CDataValues);
              if Keywords<>'' then // duplicate
                Error_handling(Strict, Error_Msg, Warning_Msg, Succes,
                               MsgDuplicateElement,[NewTag,Tag]);
              if Succes then
                Keywords:=ALUTF8Encode(NewContent, CodePage);
            end;
            7 : begin // bounds
              if DefBounds then
                Error_handling(Strict, Error_Msg, Warning_Msg, Succes,
                                 MsgDuplicateElement,[NewTag,Tag]);
              if Succes and (NewContent<>'') then
                Error_handling(Strict, Error_Msg, Warning_Msg, Succes,
                                  MsgUnwantedElementContent,[NewTag,NewContent]);
              if Succes then
              begin
                lat_min_found := false;
                lat_max_found := false;
                lon_min_found := false;
                lon_max_found := false;
                repeat
                  GetFirstValue(NewHeader, bName, bValue, bEmpty, Succes);
                  ReplaceCData(bValue, CDataValues);
                  if not Succes then
                    Error_Msg:=Format(MsgInvalidElementHeader,[NewTag]);
                  if Succes and not bEmpty then
                  begin
                    case StringIndex(bName, ['minlat', 'minlon', 'maxlat', 'maxlon']) of
                      0 : begin // minlat
                        if lat_min_found then
                          Error_handling(Strict, Error_Msg, Warning_Msg, Succes,
                                  MsgDuplicateField,[bName, NewTag]);
                        if Succes then
                        begin
                          val(bValue,TheBounds.LatMin,Erreur);
                          if Erreur<>0 then
                          begin
                            Succes:=false;
                            Error_Msg:=Format(MsgInvalidAngle,[bValue,NewTag]);
                          end;
                          lat_min_found:=true;
                        end;
                      end;
                      1 : begin // minlon
                        if lon_min_found then
                          Error_handling(Strict, Error_Msg, Warning_Msg, Succes,
                                  MsgDuplicateField,[bName, NewTag]);
                        if Succes then
                        begin
                          val(bValue,TheBounds.LongMin,Erreur);
                          if Erreur<>0 then
                          begin
                            Succes:=false;
                            Error_Msg:=Format(MsgInvalidAngle,[bValue,NewTag]);
                          end;
                          lon_min_found:=true;
                        end;
                      end;
                      2 : begin // maxlat
                        if lat_max_found then
                          Error_handling(Strict, Error_Msg, Warning_Msg, Succes,
                                  MsgDuplicateField,[bName, NewTag]);
                        if Succes then
                        begin
                          val(bValue,TheBounds.LatMax,Erreur);
                          if Erreur<>0 then
                          begin
                            Succes:=false;
                            Error_Msg:=Format(MsgInvalidAngle,[bValue,NewTag]);
                          end;
                          lat_max_found:=true;
                        end;
                      end;
                      3 : begin // maxlon
                        if lon_max_found then
                          Error_handling(Strict, Error_Msg, Warning_Msg, Succes,
                                  MsgDuplicateField,[bName, NewTag]);
                        if Succes then
                        begin
                          val(bValue,TheBounds.LongMax,Erreur);
                          if Erreur<>0 then
                          begin
                            Succes:=false;
                            Error_Msg:=Format(MsgInvalidAngle,[bValue,NewTag]);
                          end;
                          lon_max_found:=true;
                        end;
                      end;
                      else
                        Error_handling(Strict, Error_Msg, Warning_Msg, Succes,
                                  MsgUnknownField,[bName, NewTag]);
                    end;
                  end;
                until (not Succes) or bEmpty;
                if Succes and not lat_min_found then
                  Error_handling(Strict, Error_Msg, Warning_Msg, Succes,
                                  MsgMissingField,['minlat', NewTag]);
                if Succes and not lon_min_found then
                  Error_handling(Strict, Error_Msg, Warning_Msg, Succes,
                                  MsgMissingField,['minlon', NewTag]);
                if Succes and not lat_max_found then
                  Error_handling(Strict, Error_Msg, Warning_Msg, Succes,
                                  MsgMissingField,['maxlat', NewTag]);
                if Succes and not lon_max_found then
                  Error_handling(Strict, Error_Msg, Warning_Msg, Succes,
                                  MsgMissingField,['maxlon', NewTag]);
                if Succes then
                  Bounds:=TheBounds;
              end;
            end;
            8 : ReadExtension(NewTag, NewHeader, NewContent, Strict, CDataValues,
                                CodePage, Error_Msg, Warning_Msg, Succes); // extensions
            else
              Error_handling(Strict, Error_Msg, Warning_Msg, Succes,
                                  MsgUnkownElement,[NewTag, Tag]);
          end;
        end;
      end;
    until (not succes) or empty;
  end;
end;

{ TGpxExtension }

procedure TGpxExtension.Analyze_XML(var header, content: string;
  const Strict: boolean; CDataValues : array of string; const CodePage : word;
  var Error_Msg, Warning_Msg: string;
  var Succes: boolean);
begin
  if Header<>'' then
    Error_handling(Strict, Error_Msg, Warning_Msg, Succes,
                               MsgUnwantedElementHeader,[Tag,Header]);
  if Succes then
    Extension_Content:=content;
end;

procedure TGpxExtension.GenerateInside(var content: TStringList;
  indent: string);
begin
  inherited;
  if Extension_Content<>'' then
    content.Add(indent+Extension_Content)
end;


procedure TGpxExtension.SetExtension_Content(const Value: string);
begin
  FExtension_Content := Value;
end;

{ TGPXBasicExt }

procedure TGPXBasicExt.DeleteExtension;
begin
 if FExtension<>nil then
 begin
   FExtension.Free;
   FExtension:=nil;
 end;
end;

procedure TGPXBasicExt.GenerateInside(var content: TStringList;
  indent: string);
begin
  inherited;
  if Extension<>nil then
    Extension.Generate(content, one_indent);
end;

procedure TGPXBasicExt.ReadExtension(NewTag, Newheader, Newcontent: string;
  const Strict: boolean; CDataValues : array of string; const CodePage : word;
  var Error_Msg, Warning_Msg: string;
  var Succes: boolean);
begin
  if NewHeader<>'' then // header not empty
    Error_handling(Strict, Error_Msg, Warning_Msg, Succes,
                               MsgUnwantedElementHeader,[NewTag,Tag]);
  if Succes and (Extension<>nil) then
  begin // duplicate
    Error_handling(Strict, Error_Msg, Warning_Msg, Succes,
                               MsgDuplicateElement,[NewTag,Tag]);
    if Succes then
      DeleteExtension;
  end;
  if Succes then
  begin
    Extension:=TGPXExtension.Create(NewTag);
    Extension.Analyze_XML(Newheader, Newcontent, Strict, CDataValues, CodePage,
                            Error_Msg, Warning_Msg, Succes);
  end;
end;

procedure TGPXBasicExt.SetExtension(const Value: TGPXExtension);
begin
  FExtension := Value;
end;

end.
