unit xml_utils;

interface

const
  one_indent : string = '  ';
  cdata_digit = 6;

function QuoteXMLString(S : string) : string;
// Quote a string so that it is complient with XML ie same type of quote (' or ")
// is not used at the same time as quoting and inside the string

procedure RemoveXMLComments(var S : string; var Succes : boolean);
// remove xml comments from the string
// succes True if successfull, False if something wrong append

procedure ReplaceCData(var S : string; CDataValues : array of string);
// find <![CDATA[..]]> if any and replace it by the real text

function ProtectXML(const S : string) : string;
// protect a string with CDATA if it contains special chars like &

procedure GetFirstElement(var S, Tag, Header, Content : string; Var Empty, Succes : boolean);
// find the first <element> in S, give the Tag, the Header and the Content
// <Tag Header>Content</Tag>
// remove the element from S
// Empty is True if nothing is found
// succes True if successfull, False if something wrong append

procedure GetFirstValue(var S, Name, Value : string; Var Empty, Succes : boolean);
// find the first value in <element> header in S, give the Name and the Value
// <Tag Name="value"...>
// remove the element from S
// Empty is True if nothing is found
// succes True if successfull, False if something wrong append

function StringIndex(const SearchString: string; StrList: array of string): Integer;
// for using "case of" with string

procedure Error_handling(const Strict: boolean; var Error_Msg, Warning_Msg: string;
  var Succes: boolean; const TheMessage : string; const Args: array of const);
// if Strict, set Succes to false and put the error message in Error_Msg
// if not Strict, juste add the error message to Warning_Msg
// error message : format(TheMessage, Args)

procedure GetTimeISO8601(S : string; Var DateTime : TDateTime; var Succes : boolean);
// read time from string in ISO 8601 format

function PosAfter(const Substr: string; S: string; StartPos : integer): Integer;
// find the first position of Substr in S, starting after StartPos


implementation

uses
  SysUtils, DateUtils;

function QuoteXMLString(S : string) : string;
// Quote a string so that it is complient with XML ie same type of quote (' or ")
// is not used at the same time as quoting and inside the string
begin
  if pos('''',S)=0 then
    result:=''''+S+'''' // no ' inside, we use it for quoting
  else
  begin // ' inside, we use " for quoting
    while pos('"',S)<>0 do
      S[pos('"',S)]:=''''; // replacement of " by ' inside the string
    result:='"'+S+'"';
  end;
end;

procedure RemoveXMLComments(var S : string; var Succes : boolean);
// remove xml comments from the string
// succes True if successfull, False if something wrong append
var
  Start, Stop : integer;
begin
  Succes:=true;
  Start:=pos('<!--',S);
  while Succes and (Start<>0) do
  begin
    Stop:=Pos('-->',S);
    if Stop=0 then
      Succes:=false // can not find the end of the comment
    else
    begin
      delete(S, Start, Stop-Start+3);
      Start:=pos('<!--',S); // looking for next comment
    end;
  end;
end;


procedure ReplaceCData(var S : string; CDataValues : array of string);
// find <![CDATA[..]]> if any and replace it by the real text
// CDATA should only be present one time except in <extension>
var
  Start, N : integer;
begin
  while pos('<![CDATA[',S)<>0 do
  begin
    Start:=pos('<![CDATA[',S);
    N:=StrToInt(copy(S,Start+9,cdata_digit));
    delete(S,Start,12+cdata_digit);
    insert(CDataValues[N], S, Start);
  end;
end;

function ProtectXML(const S : string) : string;
// protect a string with CDATA if it contains special chars like &
begin
  if (pos('&',S)<>0) or (pos('<',S)<>0) or (pos('>',S)<>0) then
    result:='<![CDATA['+S+']]>'
  else
    result:=S;
end;


procedure GetFirstElement(var S, Tag, Header, Content : string; Var Empty, Succes : boolean);
// find the first <element> in S, give the Tag, the Header and the Content
// <Tag Header>Content</Tag>
// could also be <Tag Header />
// remove the element from S
// Empty is True if nothing is found
// succes True if successfull, False if something wrong append
var
  Start, Stop : integer;
begin
  Succes:=true;
  Empty:=false;
  while (length(S)>0) and (S[1]=' ') do
    delete(S,1,1);
  if length(S)=0 then
    Empty:=true
  else
  begin
    if S[1]<>'<' then
      Succes:=false // the < for opening the element not found
    else
    begin
      Stop:=pos('>',S);
      if Stop=0 then
        Succes:= false // the > for closing the header not found
      else
      begin
        while S[Stop-1]=' ' do // removing spaces before >
        begin
          delete(S, Stop-1, 1);
          dec(Stop);
        end;
        Start:=pos(' ',S);
        if (Start=0) or (Start>Stop) then
        begin // nothing inside the opening part i.e. <Tag> only
          Tag:=copy(S,2,Stop-2);
          Header:='';
        end
        else // something inside the opening part i.e. <Tag field="value"...>
        begin
          Tag:=copy(S,2,Start-2);
          Header:=copy(S,Start+1,Stop-Start-1);
        end;
        delete(S,1, Stop);
        if (Header<>'') and (Header[length(Header)]='/') then
        begin // self-closing element <element/>
          delete(Header,length(Header),1);
          Content:='';
        end
        else
        begin // looking for closing part </element> or </element >
          Start:=pos('</'+Tag+'>',S); // for </element>
          if Start=0 then
          begin
            Start:=pos('</'+Tag+' ',S); // for </element >
            if Start=0 then
              Succes:=false // can not find the closing part of the element i.e. </tag>
          end;
          if Succes then
          begin
            Content:=copy(S,1,Start-1);
            delete(S,1,Start+length(Tag)+1);
            while (length(S)>0) and (S[1]=' ') do // removing spaces in </element   >
              delete(S,1,1);
            if (length(S)=0) or (S[1]<>'>') then
              Succes:=false // the > for closing the end not found
            else
              delete(S,1,1); // removing the last >
          end;
        end;  
      end;
    end;
  end;
end;

procedure GetFirstValue(var S, Name, Value : string; Var Empty, Succes : boolean);
// find the first value in <element> header in S, give the Name and the Value
// <Tag Name="value"...>
// remove the element from S
// Empty is True if nothing is found
// succes True if successfull, False if something wrong append
var
  Stop : integer;
begin
  Succes:=true;
  Empty:=false;
  while (length(S)>0) and (S[1]=' ') do
    delete(S,1,1);
  if length(S)=0 then
    Empty:=true
  else
  begin
    Stop:=pos('=',S);
    if Stop=0 then
      Succes:=false // no = found, no value???
    else
    begin
      Name:=copy(S,1,Stop-1);
      while Name[length(Name)]=' ' do // removing spaces between the name and =
        delete(Name,length(Name),1);
      delete(S,1,Stop);
      while (length(S)>0) and (S[1]=' ') do
        delete(S,1,1);
      if length(S)=0 then
        Succes:=false // no value found after =
      else
      begin
        case S[1] of
          '''' : begin
            delete(S,1,1);
            Stop:=pos('''',S);
            if Stop=0 then
              succes:=false // closing of quoting not found
            else
            begin
              Value:=copy(S,1,Stop-1);
              delete(S, 1, Stop);
            end;
          end;
          '"' : begin
            delete(S,1,1);
            Stop:=pos('"',S);
            if Stop=0 then
              succes:=false // closing of quoting not found
            else
            begin
              Value:=copy(S,1,Stop-1);
              delete(S, 1, Stop);
            end;
          end;
        else
          Succes:=false; // value not quoted with ' or "
        end;
      end;
    end;
  end;
end;

function StringIndex(const SearchString: string; StrList: array of string): Integer;
{var
  I: Integer; }
begin
  result:=high(StrList);
  while (result>=0) and (SearchString<>StrList[result]) do
    dec(result);
{  Result:= -1;
  for I:= 0 to High(StrList) do
    if SearchString = StrList[I] then
    begin
      Result:= I;
      Break;
    end; }
end;

procedure Error_handling(const Strict: boolean; var Error_Msg, Warning_Msg: string;
  var Succes: boolean; const TheMessage : string; const Args: array of const);
begin
  if Strict then
  begin
    Succes:=false;
    Error_Msg:=Format(TheMessage,Args);
  end
  else
    Warning_Msg:=Warning_Msg+#13#10+Format(TheMessage,Args);
end;

procedure GetTimeISO8601(S : string; Var DateTime : TDateTime; var Succes : boolean);
// read time from string in ISO 8601 format
var
  DayPart, TimePart, Fuseau : string;
  Date, Time : TDateTime;
begin
  Succes:=true;
  try
    if Pos('-',S) in [1..8] then
    begin  // converting from extended notation to compact one
      while Pos('-',S) in [1..8] do // (<=8 not to suppress the - from time zone)
        delete(S,Pos('-',S),1);
      while Pos(':',S)<>0 do
        delete(S,Pos(':',S),1);
      while Pos(',',S)<>0 do
        delete(S,Pos(',',S),1);
      while Pos('.',S)<>0 do
        delete(S,Pos('.',S),1);
    end;
    DayPart:=copy(S,1,pos('T',S)-1);
    TimePart:=copy(S,pos('T',S)+1,length(S));
    Fuseau:='';

    // looking for time zone
    if TimePart[length(TimePart)]='Z' then
      delete(TimePart,length(TimePart),1)
    else
    begin
      if pos('+',TimePart)<>0 then
      begin
        Fuseau:=copy(TimePart, pos('+',TimePart), 5);
        delete(TimePart, pos('+', TimePart),5);
      end
      else
      begin
        if pos('-',TimePart)<>0 then
        begin
          Fuseau:=copy(TimePart, pos('-',TimePart), 5);
          delete(TimePart, pos('-', TimePart),5);
        end
      end;
    end;

    // managing date part
    if DayPart[5]='W' then
    // YYYY Www D
      Date := EncodeDateWeek(StrToInt(copy(DayPart,1,4)), StrToInt(DayPart[6]+DayPart[7]),
                                        StrToInt(DayPart[8]))
    else
    begin
      if length(DayPart)=8 then
        // YYYY MM DD
        Date := EncodeDate(StrToInt(copy(DayPart,1,4)), StrToInt(DayPart[5]+DayPart[6]),
                                        StrToInt(DayPart[7]+DayPart[8]))
      else // YYYY DDD
        Date := EncodeDateDay(StrToInt(copy(DayPart,1,4)),StrToInt(DayPart[5]+DayPart[6]+DayPart[7]));
    end;

    // managing time part
    case length(TimePart) of
      2 : Time := EncodeTime(StrToInt(TimePart[1]+TimePart[2]),0,0,0);
      4 : Time := EncodeTime(StrToInt(TimePart[1]+TimePart[2]),StrToInt(TimePart[3]+TimePart[4]),0,0);
      6 : Time := EncodeTime(StrToInt(TimePart[1]+TimePart[2]),StrToInt(TimePart[3]+TimePart[4]),
                      StrToInt(TimePart[5]+TimePart[6]),0);
      7 : Time := EncodeTime(StrToInt(TimePart[1]+TimePart[2]),StrToInt(TimePart[3]+TimePart[4]),
           StrToInt(TimePart[5]+TimePart[6]),100*StrToInt(TimePart[7]));
      8 : Time := EncodeTime(StrToInt(TimePart[1]+TimePart[2]),StrToInt(TimePart[3]+TimePart[4]),
           StrToInt(TimePart[5]+TimePart[6]),10*StrToInt(TimePart[7]+TimePart[8]));
      else
        Time := EncodeTime(StrToInt(TimePart[1]+TimePart[2]),StrToInt(TimePart[3]+TimePart[4]),
           StrToInt(TimePart[5]+TimePart[6]),StrToInt(TimePart[7]+TimePart[8]+TimePart[9]));
    end;
    DateTime:=Date+Time;

    // managing time zone
    if Fuseau<>'' then
    begin
      if length(Fuseau)=3 then
        DateTime:=IncHour(DateTime,-StrToInt(Fuseau))
      else
      begin
        DateTime:=IncHour(DateTime,-StrToInt(Fuseau[1]+Fuseau[2]+Fuseau[3]));
        DateTime:=IncMinute(DateTime,-StrToInt(Fuseau[1]+Fuseau[4]+Fuseau[5]));
      end;
    end;

  except
    Succes:=false;
  end;
end;

function PosAfter(const Substr: string; S: string; StartPos : integer): Integer;
// find the first position of Substr in S, starting after StartPos
begin
  delete(S, 1, StartPos-1);
  result:=pos(Substr,S);
  if result<>0 then
    inc(result, StartPos);
end;

end.
